from pathlib import Path
from importlib import resources
import inspect
import json
import pathlib
import warnings
import copy
import numpy as np
from photompy import read_ies_data, write_ies_data, total_optical_power
from .spectrum import Spectrum
from .lamp_surface import LampSurface
from .lamp_plotter import LampPlotter
from .trigonometry import to_cartesian, to_polar, attitude
from ._data import get_tlvs

VALID_LAMPS = [
    "beacon",
    "lumenizer_zone",
    "sterilray",
    "ushio_b1",
    "ushio_b1.5",
    "uvpro222_b1",
    "uvpro222_b2",
    "visium",
]

KRCL_KEYS = [
    "krypton chloride",
    "krypton-chloride",
    "krypton_chloride",
    "krcl",
    "kr-cl",
    "kr cl",
]

LPHG_KEYS = [
    "low pressure mercury",
    "low-pressure mercury",
    "mercury",
    "lphg",
    "lp-hg",
    "lp hg",
]


class Lamp:
    """
    Represents a lamp with properties defined by a photometric data file.
    This class handles the loading of IES photometric data, orienting the lamp in 3D space,
    and provides methods for moving, rotating, and aiming the lamp.

    Arguments
    -------------------
    lamp_id: str
        A unique identifier for the lamp object.
    name: str, default=None
        Non-unique display name for the lamp. If None set by lamp_id
    filename: Path, str
        If None or not pathlike, `filedata` must not be None
    filedata: Path or bytes, default=None
        Set by `filename` if filename is pathlike.
    x, y, z: floats, default=[0,0,0]
        Sets initial position of lamp in cartesian space
    angle: float, default=0
        Sets lamps initial rotation on its own axis.
    aimx, aimy, aimz: floats, default=[0,0,z-1]
        Sets initial aim point of lamp in cartesian space.
    guv_type: str
        Optional label for type of GUV source. Presently available:
        ["Krypton chloride (222 nm)", "Low-pressure mercury (254 nm)", "Other"]
    wavelength: float
        Optional label for principle GUV wavelength. Set from guv_type if guv_type
        is not "Other".
    spectra_source: Path or bytes, default=None
        Optional. Data source for spectra. May be a filepath, a binary stream,
        or a dict where the first value contains values of wavelengths, and
        the second value contains values of relative intensity.
    width, length, depth: floats, default=[None, None]
        x-axis and y-axis source extent, plus fixture depth in the units
        provided. If not provided, will be read from the .ies file. Note that
        the ies file interface labels `depth` as `height` instead.
    units: str or int in [1, 2] or None
        `feet` or `meters`. 1 corresponds to feet, 2 to `meters`. If not
        provided, will be read from .ies file, and lengt and width parameters
        will be ignored.
    source_density: int or float, default=1
        parameter that determines the fineness of the source discretization.
        Grid size follows fibonacci sequence. For an approximately square
        source, SD=1 => 1x1 grid, SD=2 => 3x3 grid, SD=3 => 5x5 grid. This is
        to ensure that a center point is always present while ensuring evenness
        of grid size.
    intensity_map: arraylike
        A relative intensity map for non-uniform sources. Must be of the same
        size as the grid generated
    enabled: bool, defualt=True
        Determines if lamp participates in calculations. A lamp may be created
        and added to a room, but disabled.

    """

    def __init__(
        self,
        lamp_id=None,
        name=None,
        filename=None,
        filedata=None,
        x=None,
        y=None,
        z=None,
        angle=None,
        aimx=None,
        aimy=None,
        aimz=None,
        intensity_units=None,
        guv_type=None,
        wavelength=None,
        spectra_source=None,
        width=None,
        length=None,
        depth=None,
        units=None,
        source_density=None,
        intensity_map=None,
        enabled=None,
    ):

        """
        TODO:
            probably __init__ needs to change initialization strategy. some kind of from_file?
            unfortunately this object is initialized with two files...maybe that is an issue
            possibly need to make all the surface emission properties into a separate class
        """

        self.lamp_id = str(lamp_id)
        self.name = lamp_id if name is None else str(name)
        self.enabled = True if enabled is None else enabled
        # position
        self.x = 0.0 if x is None else x
        self.y = 0.0 if y is None else y
        self.z = 0.0 if z is None else z
        self.position = np.array([self.x, self.y, self.z])
        # orientation
        self.angle = 0.0 if angle is None else angle
        self.aimx = self.x if aimx is None else aimx
        self.aimy = self.y if aimy is None else aimy
        self.aimz = self.z - 1.0 if aimz is None else aimz
        self.aim_point = np.array([self.aimx, self.aimy, self.aimz])

        # mW/sr or uW/cm2 typically; not directly specified in .ies file and they can vary for GUV fixtures
        self.intensity_units = self._set_intensity_units(intensity_units)

        # source type & wavelength
        self.guv_type = guv_type
        if guv_type is not None:
            if any([key in guv_type.lower() for key in KRCL_KEYS]):
                self.wavelength = 222
            elif any([key in guv_type.lower() for key in LPHG_KEYS]):
                self.wavelength = 254
            else:  # set from wavelength if guv_type not recognized
                self.wavelength = wavelength
        else:
            self.wavelength = wavelength
        if self.wavelength is not None:
            if not isinstance(self.wavelength, (int, float)):
                msg = f"Wavelength must be int or float, not {type(self.wavelength)}"
                raise TypeError(msg)

        # source values
        self.surface = LampSurface(
            width=width,
            length=length,
            depth=depth,
            units=units,
            source_density=source_density,
            intensity_map=intensity_map,
            position=self.position,
            aim_point=self.aim_point,
            angle=self.angle,
        )

        # update heading and bank
        self.aim(self.aimx, self.aimy, self.aimz)

        # spectra
        self.spectra_source = spectra_source
        self.spectra = self._load_spectra(spectra_source)

        # plotting
        self.plotter = LampPlotter(self)

        # load file and coordinates
        self.filename = filename
        self.filedata = filedata
        self._check_filename()

        # values that will be filled in if there is filedata available
        self.lampdict = None
        self.thetas = None
        self.phis = None
        self.values = None
        self.coords = None
        self.photometric_coords = None
        self.calc_state = None

        # filename is just a label, filedata controls everything.
        if self.filedata is not None:
            self._load()
            self._orient()

    @classmethod
    def from_dict(cls, data):
        """initialize class from dict"""
        keys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        if data["spectra"] is not None:
            data["spectra_source"] = {}
            for k, v in data["spectra"].items():
                if isinstance(v, str):
                    lst = list(map(float, v.split(", ")))
                elif isinstance(v, list):
                    lst = v
                data["spectra_source"][k] = np.array(lst)
        return cls(**{k: v for k, v in data.items() if k in keys})

    @classmethod
    def from_keyword(
        cls,
        key,
        lamp_id=None,
        name=None,
        x=None,
        y=None,
        z=None,
        angle=None,
        aimx=None,
        aimy=None,
        aimz=None,
        intensity_units=None,
        guv_type=None,
        wavelength=None,
        width=None,
        length=None,
        depth=None,
        units=None,
        source_density=None,
        intensity_map=None,
        enabled=None,
    ):

        if key.lower() in VALID_LAMPS:
            path = "guv_calcs.data.lamp_data"
            fname = key.lower() + ".ies"
            spec_fname = key.lower() + ".csv"
            filename = resources.files(path).joinpath(fname)
            spectra_source = resources.files(path).joinpath(spec_fname)
        else:
            msg = f"{key} is not a valid lamp key. Photometric and spectral data will not be loaded."
            warnings.warn(msg, stacklevel=3)
            filename = None
            spectra_source = None

        return cls(
            lamp_id=key if lamp_id is None else lamp_id,
            name=name,
            filename=filename,
            x=x,
            y=y,
            z=z,
            angle=angle,
            aimx=aimx,
            aimy=aimy,
            aimz=aimz,
            intensity_units=intensity_units,
            guv_type=guv_type,
            wavelength=wavelength,
            spectra_source=spectra_source,
            width=width,
            length=length,
            depth=depth,
            units=units,
            source_density=source_density,
            intensity_map=intensity_map,
            enabled=enabled,
        )

    def move(self, x=None, y=None, z=None):
        """Designate lamp position in cartesian space"""
        # determine new position   selected_lamp.
        x = self.x if x is None else x
        y = self.y if y is None else y
        z = self.z if z is None else z
        position = np.array([x, y, z])
        # update aim point based on new position
        diff = position - self.position
        self.aim_point += diff
        self.aimx, self.aimy, self.aimz = self.aim_point
        # update position
        self.position = position
        self.x, self.y, self.z = self.position

        self.surface.set_orientation(
            mounting_position=self.position, aim_point=self.aim_point
        )
        return self

    def rotate(self, angle):
        """designate lamp orientation with respect to its z axis"""
        self.angle = angle
        self.surface.set_angle(angle)
        return self

    def aim(self, x=None, y=None, z=None):
        """aim lamp at a point in cartesian space"""
        x = self.aimx if x is None else x
        y = self.aimy if y is None else y
        z = self.aimz if z is None else z
        self.aim_point = np.array([x, y, z])
        self.aimx, self.aimy, self.aimz = self.aim_point
        xr, yr, zr = self.aim_point - self.position
        self.heading = np.degrees(np.arctan2(yr, xr))
        self.bank = np.degrees(np.arctan2(np.sqrt(xr ** 2 + yr ** 2), zr) - np.pi)
        # update grid points
        self.surface.set_orientation(
            mounting_position=self.position, aim_point=self.aim_point
        )
        return self

    def transform(self, coords, scale=1):
        """
        Transforms the given coordinates based on the lamp's orientation and position.
        Applies rotation, then aiming, then scaling, then translation.
        Scale parameter should generally only be used for photometric_coords
        """
        coords = np.array(attitude(coords.T, roll=0, pitch=0, yaw=self.angle)).T
        coords = np.array(
            attitude(coords.T, roll=0, pitch=self.bank, yaw=self.heading)
        ).T
        coords = (coords.T / scale).T + self.position
        return coords

    def set_orientation(self, orientation, dimensions=None, distance=None):
        """
        set orientation/heading.
        alternative to setting aim point with `aim`
        distinct from rotation; applies to a tilted lamp. to rotate a lamp along its axis,
        use the `rotate` method
        """
        # orientation = (orientation + 360) % 360
        self.heading = orientation
        self._recalculate_aim_point(dimensions=dimensions, distance=distance)
        return self

    def set_tilt(self, tilt, dimensions=None, distance=None):
        """
        set tilt/bank
        alternative to setting aim point with `aim`
        """
        # tilt = (tilt + 360) % 360
        self.bank = tilt
        self._recalculate_aim_point(dimensions=dimensions, distance=distance)
        return self

    def set_source_density(self, source_density):
        """change source discretization"""
        self.surface.set_source_density(source_density)

    def set_units(self, units):
        """set"""
        self.surface.set_units(units)
        return self

    def set_width(self, width):
        """change x-axis extent of lamp emissive surface"""
        self.surface.set_width(width)
        return self

    def set_length(self, length):
        """change y-axis extent of lamp emissive surface"""
        self.surface.set_length(length)
        return self

    def set_depth(self, depth):
        """change the z-axis offset of where the lamp's emissive surface is"""
        self.surface.set_depth(depth)

    def get_calc_state(self):
        """
        return a set of paramters that, if changed, indicate that
        this lamp must be recalculated
        """
        # this needs summed to make comparison not fail, might need to investigate later
        intensity_map_orig = (
            self.surface.intensity_map_orig.sum()
            if self.surface.intensity_map_orig is not None
            else None
        )
        return [
            self.filedata,
            self.x,
            self.y,
            self.z,
            self.angle,
            self.aimx,
            self.aimy,
            self.aimz,
            self.surface.length,  # only for nearfield
            self.surface.width,  # ""
            self.surface.depth,
            self.surface.units,  # ""
            self.surface.source_density,  # ""
            intensity_map_orig,
        ]

    def get_update_state(self):
        return [self.intensity_units]

    def get_total_power(self):
        """return the lamp's total optical power"""
        return total_optical_power(self.lampdict["interp_vals"])

    def get_limits(self, standard=0):
        """
        get the threshold limit values for this lamp. Returns tuple
        (skin_limit, eye_limit) Will use the lamp spectrum if provided;
        if not provided will use wavelength; if neither is defined, returns
        (None, None). Standard may be a string in:
            [`ANSI IES RP 27.1-22`, `IEC 62471-6:2022`]
        Or an integer corresponding to the index of the desired standard.
        """
        if self.spectra is not None:
            skin_tlv, eye_tlv = get_tlvs(self.spectra, standard)
        elif self.wavelength is not None:
            skin_tlv, eye_tlv = get_tlvs(self.wavelength, standard)
        else:
            skin_tlv, eye_tlv = None, None
        return skin_tlv, eye_tlv

    def get_cartesian(self, scale=1, sigfigs=9):
        """Return lamp's true position coordinates in cartesian space"""
        return self.transform(self.coords, scale=scale).round(sigfigs)

    def get_polar(self, sigfigs=9):
        """Return lamp's true position coordinates in polar space"""
        cartesian = self.transform(self.coords) - self.position
        return np.array(to_polar(*cartesian.T)).round(sigfigs)

    def reload(self, filename=None, filedata=None):
        """
        replace the ies file without erasing any position/rotation/eing information
        can be used to load an ies file after initialization
        """

        self.filename = filename
        self.filedata = filedata
        # if filename is a path, filedata is filename
        self._check_filename

        if self.filedata is not None:
            self._load()
            self._orient()
        else:
            self.lampdict = None
            self.thetas = None
            self.phis = None
            self.values = None
            self.coords = None
            self.photometric_coords = None
            self.spectra = None

    def load_spectra(self, spectra_source):
        """external method for loading spectra after lamp object has been instantiated"""
        self.spectra = self._load_spectra(spectra_source)

    def load_intensity_map(self, intensity_map):
        """external method for loading relative intensity map after lamp object has been instantiated"""
        self.surface.load_intensity_map(intensity_map)

    def copy(self, lamp_id):
        """copy the lamp object with a new ID"""
        lamp = copy.deepcopy(self)
        lamp.lamp_id = lamp_id
        return lamp

    def save_ies(self, fname=None, original=False):
        """
        Save the current lamp paramters as an .ies file; alternatively, save the original ies file.
        """
        if self.lampdict is not None:
            if original:
                iesbytes = write_ies_data(self.lampdict, valkey="original_vals")
            else:
                updated_lampdict = self.get_lampdict()
                iesbytes = write_ies_data(updated_lampdict, valkey="original_vals")

            # write to file if provided, otherwise
            if fname is not None:
                with open(fname, "wb") as file:
                    file.write(iesbytes)
            else:
                return iesbytes
        else:
            return None

    def save_lamp(self, filename=None):
        """
        save just the minimum number of parameters required to re-instantiate the lamp
        Returns dict. If filename is not None, saves dict as json.
        """

        data = {}
        data["lamp_id"] = self.lamp_id
        data["name"] = self.name
        data["x"] = self.x
        data["y"] = self.y
        data["z"] = self.z
        data["angle"] = self.angle
        data["aimx"] = self.aimx
        data["aimy"] = self.aimy
        data["aimz"] = self.aimz
        data["intensity_units"] = self.intensity_units
        data["guv_type"] = self.guv_type
        data["wavelength"] = self.wavelength
        data["width"] = self.surface.width
        data["length"] = self.surface.length
        data["depth"] = self.surface.depth
        data["units"] = self.surface.units
        data["source_density"] = self.surface.source_density

        if self.surface.intensity_map_orig is None:
            data["intensity_map"] = None
        else:
            data["intensity_map"] = self.surface.intensity_map_orig.tolist()

        data["enabled"] = True

        data["filename"] = self.filename
        filedata = self.save_ies()
        data["filedata"] = filedata.decode() if filedata is not None else None

        if self.spectra is not None:
            spectra_dict = self.spectra.to_dict(as_string=True)
            keys = list(spectra_dict.keys())[0:2]  # keep the first two keys only
            data["spectra"] = {key: spectra_dict[key] for key in keys}
        else:
            data["spectra"] = None

        if filename is not None:
            with open(filename, "w") as json_file:
                json.dump(data, json_file, indent=4)

        return data

    def plot_ies(self, **kwargs):
        """see LampPlotter.plot_ies"""
        return self.plotter.plot_ies(**kwargs)

    def plot_web(self, **kwargs):
        """see LampPlotter.plot_web"""
        return self.plotter.plot_web(**kwargs)

    def plot_3d(self, **kwargs):
        """see LampPlotter.plot_3d"""
        return self.plotter.plot_3d(**kwargs)

    def plot_spectra(self, **kwargs):
        """see LampPlotter.plot_spectra and Spectrum.plot"""
        return self.plotter.plot_spectra(**kwargs)

    def plot_surface(self, **kwargs):
        """see LampSurface.plot_surface"""
        return self.surface.plot_surface(**kwargs)

    def _check_filename(self):
        """
        determine datasource
        if filename is a path AND file exists AND filedata is None, it replaces filedata
        otherwise filedata stays the same
        """
        FILE_IS_PATH = False
        # if filename is string, check if it's a path
        if isinstance(self.filename, (str, pathlib.PosixPath)):
            if Path(self.filename).is_file():
                FILE_IS_PATH = True
            else:
                if self.filedata is None:
                    warnings.warn(
                        f"File {self.filename} not found. Provide a valid file or the filedata."
                    )
        # if filename is a path and exists, it will replace filedata, but only if filedata wasn't specified to begin with
        if FILE_IS_PATH and self.filedata is None:
            self.filedata = Path(self.filename).read_text()

    def _load(self):
        """
        Loads lamp data from an IES file and initializes photometric properties.
        """

        self.lampdict = read_ies_data(self.filedata)
        self.thetas = self.lampdict["full_vals"]["thetas"]
        self.phis = self.lampdict["full_vals"]["phis"]
        self.values = self.lampdict["full_vals"]["values"]
        self.surface.update_from_lampdict(self.lampdict)

    def get_lampdict(self):
        """
        update the lampdict object for writing a new ies file
        TODO: maybe include multiplier somehow?
        """

        lampdict = self.lampdict.copy()
        if self.surface.units == "feet":
            lampdict["units_type"] = 1
        elif self.surface.units == "meters":
            lampdict["units_type"] = 2

        lampdict["width"] = self.surface.width
        lampdict["length"] = self.surface.length
        lampdict["height"] = self.surface.depth
        return lampdict

    def _orient(self):
        """
        Initializes the orientation of the lamp based on its photometric data.
        """

        # true value coordinates
        tgrid, pgrid = np.meshgrid(self.thetas, self.phis)
        tflat, pflat = tgrid.flatten(), pgrid.flatten()
        tflat = 180 - tflat  # to account for reversed z direction
        x, y, z = to_cartesian(tflat, pflat, 1)
        self.coords = np.array([x, y, z]).T

        # photometric web coordinates
        xp, yp, zp = to_cartesian(tflat, pflat, self.values.flatten())
        self.photometric_coords = np.array([xp, yp, zp]).T

    def _recalculate_aim_point(self, dimensions=None, distance=None):
        """
        internal method to call if setting tilt/bank or orientation/heading
        if `dimensions` is passed, `distance` is not used
        """
        distance = 1 if distance is None else distance
        heading_rad = np.radians(self.heading)
        # Correcting bank angle for the pi shift
        bank_rad = np.radians(self.bank - 180)

        # Convert from spherical to Cartesian coordinates
        dx = np.sin(bank_rad) * np.cos(heading_rad)
        dy = np.sin(bank_rad) * np.sin(heading_rad)
        dz = np.cos(bank_rad)
        if dimensions is not None:
            distances = []
            dimx, dimy, dimz = dimensions
            if dx != 0:
                distances.append((dimx - self.x) / dx if dx > 0 else self.x / -dx)
            if dy != 0:
                distances.append((dimy - self.y) / dy if dy > 0 else self.y / -dy)
            if dz != 0:
                distances.append((dimz - self.z) / dz if dz > 0 else self.z / -dz)
            distance = min([d for d in distances])
        self.aim_point = self.position + np.array([dx, dy, dz]) * distance
        self.aimx, self.aimy, self.aimz = self.aim_point
        self.surface.set_orientation(
            mounting_position=self.position, aim_point=self.aim_point
        )

    def _set_intensity_units(self, arg):
        """determine the units of the radiant intensity"""
        if arg is not None:
            msg = f"Intensity unit {arg} not recognized. Using default value `mW/sr`"
            if isinstance(arg, int):
                if arg == 0:
                    intensity_units = "mW/sr"
                elif arg == 1:
                    intensity_units = "uW/cm2"
                else:
                    warnings.warn(msg, stacklevel=3)
                    intensity_units = "mW/sr"
            elif isinstance(arg, str):
                if arg.lower() in ["mw/sr", "uw/cm2", "uw/cm²"]:
                    intensity_units = arg
                else:
                    warnings.warn(msg, stacklevel=3)
                    intensity_units = "mW/sr"
            else:
                msg = f"Datatype {type(arg)} for intensity units not recognized. Using default value `mW/Sr`"
                intensity_units = "mW/sr"
        else:
            intensity_units = "mW/sr"
        return intensity_units

    def _load_spectra(self, spectra_source):
        """initialize a Spectrum object from the source"""
        if isinstance(spectra_source, dict):
            spectra = Spectrum.from_dict(spectra_source)
        elif isinstance(spectra_source, (str, pathlib.Path, bytes)):
            spectra = Spectrum.from_file(spectra_source)
        elif isinstance(spectra_source, tuple):
            spectra = Spectrum(spectra_source[0], spectra_source[1])
        elif spectra_source is None:
            spectra = None
        else:
            spectra = None
            msg = f"Datatype {type(spectra_source)} not recognized spectral data source"
            warnings.warn(msg, stacklevel=3)
        return spectra



import inspect
import warnings
import json
import copy
import numpy as np
import matplotlib.pyplot as plt
from .calc_manager import LightingCalculator
from ._helpers import rows_to_bytes


class CalcZone(object):
    """
    Base class representing a calculation zone.

    This class provides a template for setting up zones within which various
    calculations related to lighting conditions are performed. Subclasses should
    provide specific implementations of the coordinate setting method.

    NOTE: I changed this from an abstract base class to an object superclass
    to make it more convenient to work with the website, but this class doesn't really
    work on its own

    Parameters:
    --------
    zone_id: str
        identification tag for internal tracking
    name: str, default=None
        externally visible name for zone
    dimensions: array of floats, default=None
        array of len 2 if CalcPlane, of len 3 if CalcVol
    offset: bool, default=True
    fov80: bool, default=False
        apply 80 degree field of view filtering - used for calculating eye limits
        Legacy property. To be removed.
    fov_vert: float
        vertical field of view filtering. For calculating eye limits
    fov_horiz: float
        horizontal field of view filtering. Useful for not double-counting lamps
        pointed in opposite direction
    vert: bool, default=False
        calculate vertical irradiance only
    horiz: bool, default=False
        calculate horizontal irradiance only
    dose: bool, default=False
        whether to calculate a dose over N hours or just fluence
    hours: float, default = 8.0
        number of hours to calculate dose over. Only relevant if dose is True.
    enabled: bool, default = True
        whether or not the calc zone is enabled for calculations
    """

    def __init__(
        self,
        zone_id=None,
        name=None,
        offset=None,
        fov80=None,  # legacy!! just here for backwards compatibility
        fov_vert=None,
        fov_horiz=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):
        self.zone_id = str(zone_id)
        self.name = zone_id if name is None else name
        self.offset = True if offset is None else offset
        if fov80 and fov_vert is None:
            fov_vert = 80  # set if legacy value is present
        self.fov_vert = 180 if fov_vert is None else fov_vert
        self.fov_horiz = 360 if fov_horiz is None else abs(fov_horiz)
        self.vert = False if vert is None else vert
        self.horiz = False if horiz is None else horiz
        self.dose = False if dose is None else dose
        if self.dose:
            self.units = "mJ/cm²"
        else:
            self.units = "uW/cm²"
        self.hours = 8.0 if hours is None else abs(hours)  # only used if dose is true
        self.enabled = True if enabled is None else enabled
        self.show_values = True if show_values is None else show_values

        self.calculator = LightingCalculator(self)

        # these will all be calculated after spacing is set, which is set in the subclass
        self.calctype = "Zone"
        self.x1 = None
        self.x2 = None
        self.y1 = None
        self.y2 = None
        self.z1 = None
        self.z2 = None
        self.height = None
        self.num_x = None
        self.num_y = None
        self.num_z = None
        self.x_spacing = None
        self.y_spacing = None
        self.z_spacing = None
        self.num_points = None
        self.xp = None
        self.yp = None
        self.zp = None
        self.coords = None
        self.values = None
        self.reflected_values = None
        self.lamp_values = {}
        self.lamp_values_base = {}
        self.calc_state = None

    def save_zone(self, filename=None):

        data = {}
        data["zone_id"] = self.zone_id
        data["name"] = self.name
        data["offset"] = self.offset
        data["fov_vert"] = self.fov_vert
        data["fov_horiz"] = self.fov_horiz
        data["vert"] = self.vert
        data["horiz"] = self.horiz
        data["dose"] = self.dose
        data["hours"] = self.hours
        data["enabled"] = self.enabled
        data["show_values"] = self.show_values
        data["x1"] = self.x1
        data["x2"] = self.x2
        # data["x_spacing"] = self.x_spacing
        data["num_x"] = self.num_x
        data["y1"] = self.y1
        data["y2"] = self.y2
        # data["y_spacing"] = self.y_spacing
        data["num_y"] = self.num_y
        if isinstance(self, CalcPlane):
            data["height"] = self.height
            data["calctype"] = "Plane"
        elif isinstance(self, CalcVol):
            data["z1"] = self.z1
            data["z2"] = self.z2
            # data["z_spacing"] = self.z_spacing
            data["num_z"] = self.num_z
            data["calctype"] = "Volume"

        if filename is not None:
            with open(filename, "w") as json_file:
                json_file.write(json.dumps(data))

        return data

    @classmethod
    def from_dict(cls, data):
        keys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        return cls(**{k: v for k, v in data.items() if k in keys})

    def set_dimensions(self, dimensions):
        raise NotImplementedError

    def set_spacing(self, spacing):
        raise NotImplementedError

    def _write_rows(self):
        raise NotImplementedError

    def set_offset(self, offset):
        if type(offset) is not bool:
            raise TypeError("Offset must be either True or False")
        self.offset = offset
        self._update

    def set_value_type(self, dose):
        """
        if true values will be in dose over time
        if false
        """
        if type(dose) is not bool:
            raise TypeError("Dose must be either True or False")

        # convert values if they need converting
        if self.values is not None:
            if dose and not self.dose:
                self.values = self.values * 3.6 * self.hours
            elif self.dose and not dose:
                self.values = self.values / (3.6 * self.hours)

        self.dose = dose
        if self.dose:
            self.units = "mJ/cm2"
        else:
            self.units = "uW/cm2"

    def set_dose_time(self, hours):
        """
        Set the time over which the dose will be calculate in hours
        """
        if type(hours) not in [float, int]:
            raise TypeError("Hours must be numeric")
        self.hours = hours

    def get_update_state(self):
        return [self.fov_vert, self.fov_horiz, self.vert, self.horiz]

    def calculate_values(self, lamps, ref_manager=None, hard=False):
        """
        Calculate all the values for all the lamps
        """

        new_calc_state = self.get_calc_state()

        # updates self.lamp_values_base and self.lamp_values
        self.base_values = self.calculator.compute(lamps=lamps, hard=hard)

        if ref_manager is not None:
            # calculate reflectance -- warning, may be expensive!
            ref_manager.calculate_reflectance(self, hard=hard)
            # add in reflected values, if applicable
            self.reflected_values = ref_manager.get_total_reflectance(self)
        else:
            self.reflected_values = np.zeros(self.num_points)

        # sum
        values = self.base_values + self.reflected_values

        # convert to dose
        if self.dose:
            values *= 3.6 * self.hours

        self.values = values
        self.calc_state = new_calc_state

        return self.values

    def export(self, fname=None):
        """
        export the calculation zone's results to a .csv file
        if the spacing has been updated but the values not recalculated,
        exported values will be blank.
        """
        try:
            rows = self._write_rows()  # implemented in subclass
            csv_bytes = rows_to_bytes(rows)

            if fname is not None:
                with open(fname, "wb") as csvfile:
                    csvfile.write(csv_bytes)
            else:
                return csv_bytes
        except NotImplementedError:
            pass

    def copy(self, zone_id):
        """
        return a copy of this CalcZone with the same attributes and a new zone_id
        """
        zone = copy.deepcopy(self)
        zone.zone_id = zone_id
        # clear calculated values
        zone.values = None
        zone.reflected_values = None
        zone.lamp_values = {}
        zone.lamp_values_base = {}
        return zone


class CalcVol(CalcZone):
    """
    Represents a volumetric calculation zone.
    A subclass of CalcZone designed for three-dimensional volumetric calculations.
    """

    def __init__(
        self,
        zone_id,
        name=None,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        z1=None,
        z2=None,
        num_x=None,
        num_y=None,
        num_z=None,
        x_spacing=None,
        y_spacing=None,
        z_spacing=None,
        offset=None,
        fov80=None,
        fov_vert=None,
        fov_horiz=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
        values=None,
    ):

        super().__init__(
            zone_id=zone_id,
            name=name,
            offset=offset,
            fov80=fov80,
            fov_vert=fov_vert,
            fov_horiz=fov_horiz,
            vert=vert,
            horiz=horiz,
            dose=dose,
            hours=hours,
            enabled=enabled,
            show_values=show_values,
        )
        self.calctype = "Volume"
        self.x1 = 0 if x1 is None else x1
        self.x2 = 6 if x2 is None else x2
        self.y1 = 0 if y1 is None else y1
        self.y2 = 4 if y2 is None else y2
        self.z1 = 0 if z1 is None else z1
        self.z2 = 2.7 if z2 is None else z2
        self.dimensions = ((self.x1, self.y1, self.z1), (self.x2, self.y2, self.z2))

        self.num_x = 30 if num_x is None else abs(num_x)
        self.num_y = 30 if num_y is None else abs(num_y)
        self.num_z = 10 if num_z is None else abs(num_z)

        default_x = abs(self.x2 - self.x1) / self.num_x
        default_y = abs(self.y2 - self.y1) / self.num_y
        default_z = abs(self.z2 - self.z1) / self.num_z

        # only allow setting by spacing if num_x / num_y / num_z are None
        if num_x is None:
            self.x_spacing = default_x if x_spacing is None else abs(x_spacing)
            self.num_x = int(abs((self.x2 - self.x1) / self.x_spacing))
        else:
            self.x_spacing = default_x
            if x_spacing is not None:
                msg = f"Passed x_spacing value will be ignored for calc zone {self.zone_id}, num_x used instead"
                warnings.warn(msg, stacklevel=3)

        if num_y is None:
            self.y_spacing = default_y if y_spacing is None else abs(y_spacing)
            self.num_y = int(abs((self.y2 - self.y1) / self.y_spacing))
        else:
            self.y_spacing = default_y
            if y_spacing is not None:
                msg = f"Passed y_spacing value will be ignored for calc zone {self.zone_id}, num_y used instead"
                warnings.warn(msg, stacklevel=3)

        if num_z is None:
            self.z_spacing = default_z if z_spacing is None else abs(z_spacing)
            self.num_z = int(abs((self.z2 - self.z1) / self.z_spacing))
        else:
            self.z_spacing = default_z
            if z_spacing is not None:
                msg = f"Passed z_spacing value will be ignored for calc zone {self.zone_id}, num_z used instead"
                warnings.warn(msg, stacklevel=3)

        self._update()
        self.values = np.zeros(self.num_points)
        self.reflected_values = np.zeros(self.num_points)

    def get_calc_state(self):
        """
        return a set of paramters that, if changed, indicate that
        this calc zone must be recalculated
        """
        return [
            self.offset,
            self.x1,
            self.x2,
            self.x_spacing,
            self.num_x,
            self.y1,
            self.y2,
            self.y_spacing,
            self.num_y,
            self.z1,
            self.z2,
            self.z_spacing,
            self.num_z,
        ]

    def set_dimensions(self, x1=None, x2=None, y1=None, y2=None, z1=None, z2=None):
        self.x1 = self.x1 if x1 is None else x1
        self.x2 = self.x2 if x2 is None else x2
        self.y1 = self.y1 if y1 is None else y1
        self.y2 = self.y2 if y2 is None else y2
        self.z1 = self.z1 if z1 is None else z1
        self.z2 = self.z2 if z2 is None else z2
        self.dimensions = ((self.x1, self.y1, self.z1), (self.x2, self.y2, self.z2))
        self._update()

    def set_spacing(self, x_spacing=None, y_spacing=None, z_spacing=None):
        """
        set the spacing desired in the dimension
        """
        self.x_spacing = self.x_spacing if x_spacing is None else abs(x_spacing)
        self.y_spacing = self.y_spacing if y_spacing is None else abs(y_spacing)
        self.z_spacing = self.z_spacing if z_spacing is None else abs(z_spacing)
        numx = int(abs(self.x2 - self.x1) / self.x_spacing)
        numy = int(abs(self.y2 - self.y1) / self.y_spacing)
        numz = int(abs(self.z2 - self.z1) / self.z_spacing)

        if numx == 0:
            msg = f"x_spacing too large. Minimum spacing:{self.x2-self.x1}"
            warnings.warn(msg, stacklevel=3)
            numx += 1
        if numy == 0:
            msg = f"y_spacing too large. Minimum spacing:{self.y2-self.y1}"
            warnings.warn(msg, stacklevel=3)
            numy += 1
        if numz == 0:
            msg = f"z_spacing too large. Minimum spacing:{self.z2-self.z1}"
            warnings.warn(msg, stacklevel=3)
            numz += 1
        self.num_x = numx
        self.num_y = numy
        self.num_z = numz
        self._update()

    def set_num_points(self, num_x=None, num_y=None, num_z=None):
        """
        set the number of points desired in a dimension, instead of setting the spacing
        """
        self.num_x = self.num_x if num_x is None else num_x
        self.num_y = self.num_y if num_y is None else num_y
        self.num_z = self.num_z if num_z is None else num_z
        if self.num_x == 0:
            warnings.warn("Number of x points must be at least 1")
            self.num_x += 1
        if self.num_y == 0:
            warnings.warn("Number of y points must be at least 1")
            self.num_y += 1
        if self.num_z == 0:
            warnings.warn("Number of z points must be at least 1")
            self.num_z += 1

        self.x_spacing = abs((self.x2 - self.x1) / round(self.num_x))
        self.y_spacing = abs((self.y2 - self.y1) / round(self.num_y))
        self.z_spacing = abs((self.z2 - self.z1) / round(self.num_z))

        self._update()
        return self

    def _update(self):
        """
        Update the number of points based on the spacing, and then the points
        """
        if self.offset:
            xmult = -1 if self.x1 > self.x2 else 1
            ymult = -1 if self.y1 > self.y2 else 1
            zmult = -1 if self.z1 > self.z2 else 1
        else:
            xmult, ymult, zmult = 0, 0, 0

        if self.x1 == self.x2:
            self.num_x = 1
        if self.y1 == self.y2:
            self.num_y = 1
        if self.z1 == self.z2:
            self.num_z = 1

        x_offset = xmult * (self.x_spacing / 2)
        y_offset = ymult * (self.y_spacing / 2)
        z_offset = zmult * (self.z_spacing / 2)

        xpoints = np.linspace(self.x1 + x_offset, self.x2 - x_offset, self.num_x)
        ypoints = np.linspace(self.y1 + y_offset, self.y2 - y_offset, self.num_y)
        zpoints = np.linspace(self.z1 + z_offset, self.z2 - z_offset, self.num_z)

        self.points = [xpoints, ypoints, zpoints]
        self.xp, self.yp, self.zp = self.points

        X, Y, Z = [
            grid.reshape(-1) for grid in np.meshgrid(*self.points, indexing="ij")
        ]
        self.coords = np.array((X, Y, Z)).T
        self.coords = np.unique(self.coords, axis=0)

        self.num_points = np.array([len(self.xp), len(self.yp), len(self.zp)])

    def _write_rows(self):
        """
        export solution to csv file
        designed to be in the same format as the Acuity Visual export
        """

        header = """Data format notes:

         Data consists of numZ horizontal grids of fluence rate values; each grid contains numX by numY points.

         numX; numY; numZ are given on the first line of data.
         The next line contains numX values; indicating the X-coordinate of each grid column.
         The next line contains numY values; indicating the Y-coordinate of each grid row.
         The next line contains numZ values; indicating the Z-coordinate of each horizontal grid.
         A blank line separates the position data from the first horizontal grid of fluence rate values.
         A blank line separates each subsequent horizontal grid of fluence rate values.

         fluence rate values are given in µW/cm²
         
         """
        lines = header.split("\n")
        rows = [[line] for line in lines]
        rows += [self.num_points]
        rows += self.points
        num_x = self.num_points[0]
        num_y = self.num_points[1]
        num_z = self.num_points[2]
        for i in range(num_z):
            rows += [""]
            if self.values is None:
                rows += [[""] * num_x] * num_y
            elif self.values.shape != (num_x, num_y, num_z):
                rows += [[""] * num_x] * num_y
            else:
                rows += self.values.T[i].tolist()
        return rows


class CalcPlane(CalcZone):
    """
    Represents a planar calculation zone.
    A subclass of CalcZone designed for two-dimensional planar calculations at a specific height.
    """

    def __init__(
        self,
        zone_id,
        name=None,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        height=None,
        ref_surface="xy",
        num_x=None,
        num_y=None,
        x_spacing=None,
        y_spacing=None,
        offset=None,
        fov80=None,
        fov_vert=None,
        fov_horiz=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):

        super().__init__(
            zone_id=zone_id,
            name=name,
            offset=offset,
            fov80=fov80,
            fov_vert=fov_vert,
            fov_horiz=fov_horiz,
            vert=vert,
            horiz=horiz,
            dose=dose,
            hours=hours,
            enabled=enabled,
            show_values=show_values,
        )
        self.calctype = "Plane"
        self.height = 1.9 if height is None else height

        self.x1 = 0 if x1 is None else x1
        self.x2 = 6 if x2 is None else x2
        self.y1 = 0 if y1 is None else y1
        self.y2 = 4 if y2 is None else y2
        self.dimensions = ((self.x1, self.y1), (self.x2, self.y2))

        if not isinstance(ref_surface, str):
            raise TypeError("ref_surface must be a string in [`xy`,`xz`,`yz`]")
        if ref_surface.lower() not in ["xy", "xz", "yz"]:
            raise ValueError("ref_surface must be a string in [`xy`,`xz`,`yz`]")

        self.ref_surface = "xy" if ref_surface is None else ref_surface.lower()

        self.num_x = 30 if num_x is None else abs(num_x)
        self.num_y = 30 if num_y is None else abs(num_y)

        default_x = abs(self.x2 - self.x1) / self.num_x
        default_y = abs(self.y2 - self.y1) / self.num_y

        # only allow setting by spacing if num_x / num_y are None
        if num_x is None:
            self.x_spacing = default_x if x_spacing is None else abs(x_spacing)
            self.num_x = int(abs((self.x2 - self.x1) / self.x_spacing))
        else:
            self.x_spacing = default_x
            if x_spacing is not None:
                msg = f"Passed x_spacing value will be ignored for calc zone {self.zone_id}, num_x used instead"
                warnings.warn(msg, stacklevel=3)
        if num_y is None:
            self.y_spacing = default_y if y_spacing is None else abs(y_spacing)
            self.num_y = int(abs((self.y2 - self.y1) / self.y_spacing))
        else:
            self.y_spacing = default_y
            if y_spacing is not None:
                msg = f"Passed y_spacing value will be ignored for calc zone {self.zone_id}, num_y used instead"
                warnings.warn(msg, stacklevel=3)

        self._update()
        self.values = np.zeros(self.num_points)
        self.reflected_values = np.zeros(self.num_points)

    def set_height(self, height):
        """set height of calculation plane. currently we only support vertical planes"""
        if type(height) not in [float, int]:
            raise TypeError("Height must be numeric")
        self.height = height
        self._update()
        return self

    def set_dimensions(self, x1=None, x2=None, y1=None, y2=None):
        """set the dimensions and update the coordinate points"""
        self.x1 = self.x1 if x1 is None else x1
        self.x2 = self.x2 if x2 is None else x2
        self.y1 = self.y1 if y1 is None else y1
        self.y2 = self.y2 if y2 is None else y2
        self.dimensions = ((self.x1, self.y1), (self.x2, self.y2))
        self._update()
        return self

    def set_spacing(self, x_spacing=None, y_spacing=None):
        """set the fineness of the grid spacing and update the coordinate points"""
        self.x_spacing = self.x_spacing if x_spacing is None else abs(x_spacing)
        self.y_spacing = self.y_spacing if y_spacing is None else abs(y_spacing)
        numx = int(abs(self.x2 - self.x1) / self.x_spacing)
        numy = int(abs(self.y2 - self.y1) / self.y_spacing)
        if numx == 0:
            msg = f"x_spacing too large. Minimum spacing:{self.x2-self.x1}"
            warnings.warn(msg, stacklevel=3)
            numx += 1
        if numy == 0:
            msg = f"y_spacing too large. Minimum spacing:{self.y2-self.y1}"
            warnings.warn(msg, stacklevel=3)
            numy += 1
        self.num_x = numx
        self.num_y = numy
        self._update()
        return self

    def set_num_points(self, num_x=None, num_y=None):
        """
        set the number of points desired in a dimension, instead of setting the spacing
        """
        self.num_x = self.num_x if num_x is None else abs(num_x)
        self.num_y = self.num_y if num_y is None else abs(num_y)
        if self.num_x == 0:
            warnings.warn("Number of x points must be at least 1")
            self.num_x += 1
        if self.num_y == 0:
            warnings.warn("Number of y points must be at least 1")
            self.num_y += 1

        self.x_spacing = abs((self.x2 - self.x1) / round(self.num_x))
        self.y_spacing = abs((self.y2 - self.y1) / round(self.num_y))

        self._update()
        return self

    def get_calc_state(self):
        """
        return a set of paramters that, if changed, indicate that
        this calc zone must be recalculated
        """
        return [
            self.offset,
            self.x1,
            self.x2,
            self.x_spacing,
            self.num_x,
            self.y1,
            self.y2,
            self.y_spacing,
            self.num_y,
            self.height,
        ]

    def _update(self):
        """
        Update the number of points based on the spacing, and then the points
        """
        if self.offset:
            xmult = -1 if self.x1 > self.x2 else 1
            ymult = -1 if self.y1 > self.y2 else 1
        else:
            xmult, ymult = 0, 0

        if self.x1 == self.x2:
            self.num_x = 1
        if self.y1 == self.y2:
            self.num_y = 1

        x_offset = xmult * (self.x_spacing / 2)
        y_offset = ymult * (self.y_spacing / 2)

        xpoints = np.linspace(self.x1 + x_offset, self.x2 - x_offset, self.num_x)
        ypoints = np.linspace(self.y1 + y_offset, self.y2 - y_offset, self.num_y)

        self.points = [xpoints, ypoints]
        self.xp, self.yp = self.points
        X, Y = [grid.reshape(-1) for grid in np.meshgrid(*self.points, indexing="ij")]

        if self.ref_surface in ["xy"]:
            Z = np.full(X.shape, self.height)
        elif self.ref_surface in ["xz"]:
            Z = Y
            Y = np.full(Y.shape, self.height)
        elif self.ref_surface in ["yz"]:
            Z = X
            X = np.full(X.shape, self.height)

        self.coords = np.stack([X.flatten(), Y.flatten(), Z.flatten()], axis=-1)

        # xy_coords = np.array([np.array((x0, y0)) for x0, y0 in zip(X, Y)])
        # zs = np.ones(xy_coords.shape[0]) * self.height

        # self.coords = np.stack([xy_coords.T[0], xy_coords.T[1], zs]).T
        # self.coords = np.unique(self.coords, axis=0)

        self.num_points = np.array([len(self.xp), len(self.yp)])

    def plot_plane(self, fig=None, ax=None, vmin=None, vmax=None, title=None):
        """Plot the image of the radiation pattern"""
        if fig is None:
            if ax is None:
                fig, ax = plt.subplots()
            else:
                fig = plt.gcf()
        else:
            if ax is None:
                ax = fig.axes[0]

        title = "" if title is None else title
        if self.values is not None:
            vmin = self.values.min() if vmin is None else vmin
            vmax = self.values.max() if vmax is None else vmax
            extent = [self.x1, self.x2, self.y1, self.y2]
            img = ax.imshow(self.values.T[::-1], extent=extent, vmin=vmin, vmax=vmax)
            cbar = fig.colorbar(img, pad=0.03)
            ax.set_title(title)
            cbar.set_label(self.units, loc="center")
        return fig, ax

    def _write_rows(self, fname=None):
        """
        export solution to csv
        """
        # if fname is None:
        # fname = self.name+".csv"
        num_x = self.num_points[0]
        num_y = self.num_points[1]

        rows = [[""] + self.points[0].tolist()]
        if self.values is None:
            vals = [[""] * num_y] * num_x
        elif self.values.shape != (num_x, num_y):
            vals = [[""] * num_y] * num_x
        else:
            vals = self.values
        rows += np.concatenate(([np.flip(self.points[1])], vals)).T.tolist()
        rows += [""]
        # zvals
        rows += [[""] + [self.height] * num_x] * num_y
        return rows



import numpy as np
import copy
from .calc_zone import CalcPlane
from .trigonometry import to_polar

class ReflectanceManager:
    """
    Class for managing reflective surfaces and their interactions

    Attributes:
    room: guv_calcs.Room
        instance of the parent class
    reflectances: dict { str: float}
        dict with keys [`floor`, `ceiling`, `south`, `north`, `east`, `west`]
        and float values between 0 and 1. All values default 0.0
    x_spacings: dict { str: float}
        dict with same keys as `reflectances` and float values greater than 0.
        All values default 0.5. Determines spacing in the relative x direction
    u_spacings: dict { str: float}
        dict with same keys as `reflectances` and float values greater than 0.
        All values default 0.5. Determines spacing in the relative y direction

    """

    def __init__(
        self,
        x,
        y,
        z,
        reflectances=None,
        x_spacings=None,
        y_spacings=None,
        max_num_passes=None,
    ):

        self.x = x
        self.y = y
        self.z = z

        keys = ["floor", "ceiling", "south", "north", "east", "west"]
        default_reflectances = {surface: 0.0 for surface in keys}
        default_spacings = {surface: 0.1 for surface in keys}

        self.reflectances = {**default_reflectances, **(reflectances or {})}
        self.x_spacings = {**default_spacings, **(x_spacings or {})}
        self.y_spacings = {**default_spacings, **(y_spacings or {})}
        self.max_num_passes = 6 if max_num_passes is None else max_num_passes

        self.zone_dict = {}  # will contain all values from all contributions
        self.surfaces = {}
        self.managers = {}
        self._initialize_surfaces()

    def set_reflectance(self, R, wall_id=None):
        """set reflectance by wall_id or, if wall_if is None, to all walls"""
        keys = self.reflectances.keys()
        if wall_id is None:
            # set this value for all walls
            for wall in keys:
                # self.reflectances[wall] = R
                self.surfaces[wall].R = R
        else:
            if wall_id not in keys:
                raise KeyError(f"wall_id must be in {keys}")
            else:
                # self.reflectances[wall_id] = R
                self.surfaces[wall_id].R = R
        # update the dict
        for key, val in self.surfaces.items():
            self.reflectances[key] = val.R

    def set_spacing(self, x_spacing=None, y_spacing=None, wall_id=None):
        """set x and y spacing by wall_id or, if wall_if is None, to all walls"""
        keys = self.x_spacings.keys()
        if wall_id is None:
            # set this value for all walls
            for wall in keys:
                self.surfaces[wall].plane.set_spacing(
                    x_spacing=x_spacing, y_spacing=y_spacing
                )
        else:
            if wall_id not in keys:
                raise KeyError(f"wall_id must be in {keys}")
            else:
                self.surfaces[wall_id].plane.set_spacing(
                    x_spacing=x_spacing, y_spacing=y_spacing
                )
        for key, val in self.surfaces.items():
            self.x_spacings[key] = val.plane.x_spacing
            self.y_spacings[key] = val.plane.y_spacing

    def _initialize_surfaces(self):

        for wall, reflectance in self.reflectances.items():
            x1, x2, y1, y2, height, ref_surface = self._get_surface_dimensions(wall)
            plane = CalcPlane(
                zone_id=wall,
                x1=x1,
                x2=x2,
                y1=y1,
                y2=y2,
                height=height,
                ref_surface=ref_surface,
                x_spacing=self.x_spacings[wall],
                y_spacing=self.y_spacings[wall],
            )
            self.surfaces[wall] = ReflectiveSurface(R=reflectance, plane=plane)

        # possibly this should go here? idk
        self.managers = {}
        for wall, surface in self.surfaces.items():
            ref_manager = copy.deepcopy(self)
            del ref_manager.surfaces[wall]
            self.managers[wall] = ref_manager

    def _get_surface_dimensions(self, wall):
        """retrieve the dimensions of a particular wall based on its id"""
        if wall == "floor":
            x1, x2, y1, y2 = 0, self.x, 0, self.y
            height = 0
            ref_surface = "xy"
        elif wall == "ceiling":
            x1, x2, y1, y2 = 0, self.x, 0, self.y
            height = self.z
            ref_surface = "xy"
        elif wall == "south":
            x1, x2, y1, y2 = 0, self.x, 0, self.z
            height = 0
            ref_surface = "xz"
        elif wall == "north":
            x1, x2, y1, y2 = 0, self.x, 0, self.z
            height = self.y
            ref_surface = "xz"
        elif wall == "west":
            x1, x2, y1, y2 = 0, self.y, 0, self.z
            height = 0
            ref_surface = "yz"
        elif wall == "east":
            x1, x2, y1, y2 = 0, self.y, 0, self.z
            height = self.x
            ref_surface = "yz"
        return x1, x2, y1, y2, height, ref_surface

    def update_dimensions(self, x=None, y=None, z=None):
        """update the wall dimensions based on changes to the Room parent class"""
        self.x = self.x if x is None else x
        self.y = self.y if y is None else y
        self.z = self.z if z is None else z
        for wall, surface in self.surfaces.items():
            x1, x2, y1, y2, height, _ = self._get_surface_dimensions(wall)
            surface.plane.height = height
            surface.plane.set_dimensions(x1, x2, y1, y2)

    def calculate_incidence(self, lamps, hard=False):
        """
        calculate the incident irradiances on all reflective walls
        """

        # first pass
        for wall, surface in self.surfaces.items():
            surface.calculate_incidence(lamps, hard=hard)
        # subsequent passes
        self._interreflectance(lamps, hard=hard)

    def _interreflectance(self, lamps, hard=False):
        """
        calculate additional interreflectance
        """
        # create dict of ref managers for each wall
        managers = self._create_managers()
        i = 0
        percent = 1  # initial
        # while i<self.max_num_passes:
        while percent > 0.01 and i < self.max_num_passes:
            pc = []
            for wall, surface in self.surfaces.items():
                init = surface.plane.values.mean()
                surface.calculate_incidence(
                    lamps, ref_manager=managers[wall], hard=hard
                )

                final = surface.plane.reflected_values.mean()
                if final != 0:
                    pc.append((abs(final - init) / final))
            if len(pc) > 0:
                percent = np.mean(pc)
            else:
                percent = 0
            managers = self._update_managers(managers)
            i = i + 1
        # print(i)

    def _create_managers(self):
        """
        create a dict of reflection managers for each wall
        """
        managers = {}
        for wall, surface in self.surfaces.items():
            ref_manager = ReflectanceManager(
                x=self.x,
                y=self.y,
                z=self.z,
                reflectances=self.reflectances,
                x_spacings=self.x_spacings,
                y_spacings=self.y_spacings,
            )
            # assign planes
            for subwall, surface in ref_manager.surfaces.items():
                new_plane = copy.deepcopy(self.surfaces[subwall].plane)
                ref_manager.surfaces[subwall].plane = new_plane
            # remove the surface being reflected upon
            del ref_manager.surfaces[wall]
            managers[wall] = ref_manager
        return managers

    def _update_managers(self, managers: dict) -> dict:
        """Update all interreflection managers with newly calculated surface incidences"""
        for wall, manager in managers.items():
            subwalls = list(manager.surfaces.keys())  # Create a static copy of keys
            for subwall in subwalls:
                # Update the values without modifying the dictionary structure
                np.copyto(
                    manager.surfaces[subwall].plane.values,
                    self.surfaces[subwall].plane.values,
                )
                new_plane = copy.deepcopy(
                    self.surfaces[subwall].plane
                )  # create new plane
                new_plane.values = (
                    new_plane.reflected_values
                )  # replace the total values with only the reflected values
                # Replace the object instead of deleting in-place
                manager.surfaces[subwall] = ReflectiveSurface(
                    R=self.surfaces[subwall].R, plane=new_plane
                )

        return managers  # Updated in place

    def calculate_reflectance(self, zone, hard=False):
        """
        calculate the reflectance contribution to a calc zone from each surface
        """

        threshold = zone.base_values.mean() * 0.01  # 1% of total value

        total_values = {}
        for wall, surface in self.surfaces.items():
            if surface.R * surface.plane.values.mean() > threshold:

                surface.calculate_reflectance(zone, hard=hard)
                values = surface.zone_dict[zone.zone_id]["values"]
                total_values[wall] = values
            else:
                total_values[wall] = np.zeros(zone.num_points)
        self.zone_dict[zone.zone_id] = total_values

        return total_values

    def get_total_reflectance(self, zone):
        """sum over all surfaces to get the total reflected values for that calc zone"""
        dct = self.zone_dict[zone.zone_id]
        values = np.zeros(zone.num_points)
        for wall, surface_vals in dct.items():
            if surface_vals is not None:
                values += surface_vals * self.reflectances[wall]
        return values


class ReflectiveSurface:
    """
    Class that represents a single reflective surface defined by a calculation
    zone and a float value R between 0 and 1.
    """

    def __init__(self, R, plane):

        if not isinstance(R, (float, int)):
            raise TypeError("R must be a float in range [0, 1]")
        if R > 1 or R < 0:
            raise ValueError("R must be a float in range [0, 1]")

        if not isinstance(plane, CalcPlane):
            raise TypeError("plane must be a CalcPlane object")

        self.R = R
        self.plane = plane
        self.max_num_passes = 0  # init
        self.zone_dict = {}

    def calculate_incidence(self, lamps, ref_manager=None, hard=False):
        """calculate incoming radiation"""
        self.plane.calculate_values(lamps=lamps, ref_manager=ref_manager, hard=hard)

    def get_calc_state(self):
        """"""
        return self.plane.get_calc_state()

    def get_update_state(self):
        """"""
        return self.plane.get_update_state() + [self.plane.values.sum()]

    def calculate_reflectance(self, zone, hard=False):
        """
        calculate the reflective contribution of this reflective surface
        to a provided calculation zone

        Arguments:
            zone: a calculation zone onto which reflectance is calculated
            lamp: optional. if provided, and incidence not yet calculated, uses this
            lamp to calculate incidence. mostly this is just for
        """
        # print(zone.zone_id, self.plane.zone_id, self.max_num_passes)
        # first make sure incident irradiance is calculated
        if self.plane.values is None:
            raise ValueError("Incidence must be calculated before reflectance")

        if self.zone_dict.get(zone.zone_id) is None:
            self.zone_dict[zone.zone_id] = {}

        calc_state = zone.get_calc_state()
        update_state = zone.get_update_state()
        surface_calc_state = self.get_calc_state()
        surface_update_state = self.get_update_state()
        NEW_ZONE = self.zone_dict[zone.zone_id].get("values") is None
        ZONE_RECALC = calc_state != self.zone_dict[zone.zone_id].get("calc_state")
        ZONE_UPDATE = update_state != self.zone_dict[zone.zone_id].get("update_state")
        SURF_RECALC = surface_calc_state != self.zone_dict[zone.zone_id].get(
            "surface_calc_state"
        )
        SURF_UPDATE = surface_update_state != self.zone_dict[zone.zone_id].get(
            "surface_update_state"
        )

        RECALCULATE = NEW_ZONE or ZONE_RECALC or SURF_RECALC or hard
        UPDATE = NEW_ZONE or ZONE_UPDATE or SURF_UPDATE or hard

        if RECALCULATE:
            distances, angles, theta = self._calculate_coordinates(zone)
        else:
            distances = self.zone_dict[zone.zone_id]["distances"]
            angles = self.zone_dict[zone.zone_id]["angles"]
            theta = self.zone_dict[zone.zone_id]["theta"]

        if UPDATE:
            I_r = self.plane.values[:, :, np.newaxis, np.newaxis, np.newaxis]
            element_size = self.plane.x_spacing * self.plane.y_spacing

            values = (I_r * abs(np.cos(angles)) * element_size) / (
                np.pi * distances ** 2
            )

            values = self._apply_filters(values, theta, zone)

            # Sum over all self.plane points to get total values at each volume point
            values = np.sum(values, axis=(0, 1))  # Collapse the dimensions
            values = values.reshape(*zone.num_points)
        else:
            values = self.zone_dict[zone.zone_id]["values"]

        # update the state
        self.zone_dict[zone.zone_id] = {
            "distances": distances,
            "angles": angles,
            "theta": theta,
            "values": values,
            "calc_state": calc_state,
            "update_state": update_state,
            "surface_calc_state": surface_calc_state,
            "surface_update_state": surface_update_state,
        }

        # if zone.zone_id not in ["floor", "ceiling", "south", "north", "east", "west"]:
        # print(self.plane.zone_id, values.mean().round(3))
        # Ensure the final array has the correct shape and return multiplied by R
        return values * self.R

    def _apply_filters(self, values, theta, zone):

        theta = theta.reshape(values.shape)

        # clean nans
        if np.isnan(values).any():
            values = np.ma.masked_invalid(values)

        # apply vertical field of view
        values[theta < 90 - zone.fov_vert / 2] = 0
        if zone.vert:
            values *= np.sin(np.radians(theta))
        if zone.horiz:
            values *= abs(np.cos(np.radians(theta)))

        return values

    def _calculate_coordinates(self, zone):
        """
        return the angles and distances between the points of the reflective
        surface and the calculation zone

        this is the expensive step!
        """
        surface_points = self.plane.coords.reshape(*self.plane.num_points, 3)
        volume_points = zone.coords.reshape(*zone.num_points, 3)

        differences = (
            volume_points - surface_points[:, :, np.newaxis, np.newaxis, np.newaxis, :]
        )
        distances = np.linalg.norm(differences, axis=-1)

        cos_theta = differences[..., 2] / distances
        angles = np.arccos(cos_theta)

        # for angle-based differences
        Theta0, Phi0, R0 = to_polar(*differences.reshape(-1, 3).T)

        return distances, angles, Theta0



import warnings
import numpy as np
from ._data import (
    get_disinfection_table,
    plot_disinfection_data,
    get_full_disinfection_table,
)


class DisinfectionCalculator:
    """
    manage calculation of disinfection rates for a Room object
    """

    def __init__(self, room):
        self.room = room

    def get_disinfection_data(self, zone_id):
        """return the dataframe and violin plot"""
        df, fluence_dict = self.get_disinfection_table(zone_id)
        fig = plot_disinfection_data(df, fluence_dict, self.room)
        return df, fig

    def get_disinfection_plot(self, zone_id):
        """
        Return a violin plot of expected disinfection rates
        """
        df, fluence_dict = self.get_disinfection_table(zone_id)
        fig = plot_disinfection_data(df, fluence_dict, self.room)
        return fig

    def get_disinfection_table(self, zone_id):
        """
        Return a table of expected disinfection rates
        """
        # check if zone_id is in room.calc_zones
        if zone_id in self.room.calc_zones.keys():
            zone = self.room.calc_zones[zone_id]
        else:
            raise KeyError(f"Calc zone {zone_id} is not in the room state.")

        fluence_dict = self._get_fluence_dict(zone)
        if len(fluence_dict) > 0:
            df = get_disinfection_table(fluence=fluence_dict, room=self.room)
            if len(fluence_dict) == 1:  # drop wavelength if unnecessary
                df = df.drop(columns="wavelength [nm]")
            # move some keys arounds
            new_keys = ["Link"] + [key for key in df.keys() if "Link" not in key]
            df = df[new_keys]
        else:
            msg = "Fluence value not available; returning full disinfection data table."
            warnings.warn(msg, stacklevel=3)
            df = get_full_disinfection_table()

        return df, fluence_dict

    def _get_fluence_dict(self, zone):
        """
        get a dict of all the wavelengths and the fluences they contribute
        to in a given calculation zone
        """
        lamp_wavelengths = self._get_lamp_wavelength_dict(zone)
        fluence_dict = {}
        for label, lamp_ids in lamp_wavelengths.items():
            vals = np.zeros(zone.values.shape)
            for lamp_id in lamp_ids:
                if lamp_id in zone.lamp_values.keys():
                    vals += zone.lamp_values[lamp_id].mean()
            fluence_dict[label] = vals.mean()
        return fluence_dict

    def _get_lamp_wavelength_dict(self, zone):
        """assign lamps to each unique wavelength contributing to the zone"""
        lamp_types = self._get_lamp_types(zone)
        if len(zone.lamp_values) == 0:
            msg = f"Calc zone {zone.zone_id} has no associated lamps."
            warnings.warn(msg)
        elif len(lamp_types) == 0:
            msg = f"Calc zone {zone.zone_id} has no associated lamps with an associated wavelength."

        lamp_wavelengths = {}
        for wavelength in lamp_types:
            val = [
                lamp.lamp_id
                for lamp in self.room.lamps.values()
                if lamp.wavelength == float(wavelength)
            ]
            lamp_wavelengths[wavelength] = val
        return lamp_wavelengths

    def _get_lamp_types(self, zone):
        """fetch a list of unique wavelengths contributing to the zone"""
        wavelengths = []
        for lamp_id in zone.lamp_values.keys():
            lamp = self.room.lamps[lamp_id]
            if lamp.wavelength is not None:
                wavelengths.append(lamp.wavelength)
            else:
                msg = f"{lamp_id} has an undefined wavelength. Its fluence contribution will not be counted."
                warnings.warn(msg, stacklevel=3)
        return np.unique(wavelengths)



from guv_calcs import *

room = Room(6,4,2.7,units='feet')
lamp =  Lamp.from_keyword("ushio_b1").move(room.x/2,room.y/2, room.z)
room.add(lamp).add_standard_zones().calculate()

print(room.calc_zones["WholeRoomFluence"].values.mean())
# from guv_calcs import *
# import numpy as np
# room = Room(x=4,y=6,z=2.7,max_num_passes=6)
# fname = "src/guv_calcs/data/lamp_data/B1 module.ies"
# # lamp1 =  Lamp("Lamp1", filename=fname).move(room.x/3,room.y/2,2.5).aim(room.x/2,room.y/2,2)
# # lamp2 =  Lamp("Lamp2", filename=fname).move(room.x*2/3,room.y/2,2.5).aim(room.x/2,room.y/2,2)
# lamp =  Lamp("Lamp", filename=fname).move(2, 0, 1.35).aim(2,6,1.35)

            # # height=0.3)
# volume_spacing = 0.15
# vol = CalcVol("RefVol",
                  # x1=0, x2=room.x, y1=0, y2=room.y, z1=0, z2=room.z,
                 # x_spacing=volume_spacing,
                 # y_spacing=volume_spacing,
                 # z_spacing=volume_spacing,
                 # )
# plane_spacing = 0.1
# plane = CalcPlane(
    # "RefPlane",
        # x1=0, x2=room.x, y1=0, y2=room.y, height=1.25,
        # x_spacing=plane_spacing,y_spacing=plane_spacing,
# )
# room.add(lamp,plane,vol)

# for R in [0,0.1,0.2,0.3,0.4,0.5,0.6,0.7,0.8,0.9]:
    # room.set_reflectance(R).calculate()
    # print(round(vol.values.mean(),2),round(plane.values.mean(),2))

# # for R in [0.1,0.9]:
    # # room.set_reflectance(R).calculate()
    # # print('R =',R,' V =',round(vol.values.mean(),2))#' P =',round(plane.values.mean(),2))
# # print('---')
# # room,vol,plane = make()
# # for R in [0.9,0.1]:
    # # room.set_reflectance(R).calculate()
    # # print('R =',R,' V =',round(vol.values.mean(),2))#,' P =',round(plane.values.mean(),2))



# # print(lamp.grid_points,lamp.photometric_distance)
# # height = 1.9
# # skin_limits = CalcPlane("SkinLimits", height = height,x1=0,x2=room.x,y1=0,y2=room.y, vert=False, horiz=False, fov80=False, dose=True)
# # eye_limits = CalcPlane("EyeLimits", height = height,x1=0,x2=room.x,y1=0,y2=room.y, vert=True, horiz=False, fov80=True, dose=True)
# # fluence = CalcVol(
        # # zone_id="WholeRoomFluence",
        # # x1=0,
        # # x2=room.x,
        # # y1=0,
        # # z1=0,
        # # z2=room.z,
    # # )
# # room.add_calc_zone(fluence)
# # room.add_calc_zone(skin_limits)
# # room.add_calc_zone(eye_limits)

# # rel_map = np.array([
	# # 0.679274542,1.004584682,0.804004854,0.758158037,0.532632147,
	# # 0.748381877,1.439455232,1.303600324,1.109088457,0.666127292,
	# # 0.768608414,1.520361381,1.498786408,1.277642934,0.679948759,
	# # 0.824905609,1.554072276,1.380461165,1.179881338,0.645563646,
	# # 0.765237325,1.183252427,1.061893204,1.007955771,0.606121899,
# # ])

# # import matplotlib.pyplot as plt
# # for val in [1,2,3]:
    # # lamp = Lamp("Lamp1", 
                # # filename="src/guv_calcs/data/lamp_data/uvpro222_b1.ies",
                # # spectra_source="src/guv_calcs/data/lamp_data/uvpro222_b1.csv",
                # # angle=270,
                # # source_density=val,
                # # intensity_map=rel_map if val==3 else None
                # # ).move(1,1,1)
    # # zone = CalcPlane("test", 
            # # height = 0.95,
            # # x1=0.9,x2=1.1,
            # # y1=0.9,y2=1.1,
            # # x_spacing=0.01,y_spacing=0.01)

    # # room.add_lamp(lamp).add_calc_zone(zone)
    # # room.calculate()
    
    # # fig,ax=zone.plot_plane()
    # # plt.show()

# # # test data retrieval functions
# # print(get_disinfection_table(fluence.values.mean()))
# # standard = get_standards()[1]
# # print(get_tlv(222,standard))

# # spectra_source="src/guv_calcs/data/lamp_data/uvpro222_b1.csv"
# # spec = Spectrum.from_file(spectra_source)
# # print(spec.get_tlv(standard))
# # print(spec.get_max_irradiance(standard))
# # print(spec.get_seconds_to_tlv(15, standard)/3600)

# # room.add_lamp(lamp).add_standard_zones().calculate()


# # room.save('tests/test.guv')
# # newroom = Room.load('tests/test.guv')


# # skin_limits.export()
# # fluence.export()

# # room.export_zip("tests/test.zip",
    # # include_plots=True,
    # # include_lamp_files=True,
    # # include_lamp_plots=True,)
    
   

# # # import matplotlib.pyplot as plt
# # # fig=skin_limits.plot_plane()
# # # plt.show()
# # lamp.plot_spectra()
# # plt.show()
# # # fig=room.plotly()
# # fig.show()

# # room.to_json('tests/test.json')
# # newroom = Room.from_json('tests/test.json')
# # newroom.calculate()




import json
import pathlib
from pathlib import Path
import numpy as np
import io
import csv


def parse_loadfile(filedata):
    """
    validate and parse a loadfile
    """

    if isinstance(filedata, (str, bytes or bytearray)):
        try:
            dct = json.loads(filedata)
        except json.JSONDecodeError:
            path = Path(filedata)
            dct = load_file(path)
    elif isinstance(filedata, pathlib.PosixPath):
        dct = load_file(filedata)

    return dct


def load_file(path):
    """load json from a"""
    if path.is_file():
        if path.suffix.lower() != ".guv":
            raise ValueError("Please provide a valid .guv file")
        with open(path, "r") as json_file:
            try:
                dct = json.load(json_file)
            except json.JSONDecodeError:
                raise ValueError(".guv file is malformed")
    else:
        raise FileNotFoundError("Please provide a valid .guv file")
    return dct


def load_csv(datasource):
    """load csv data from either path or bytes"""
    if isinstance(datasource, (str, pathlib.PosixPath)):
        filepath = Path(datasource)
        filetype = filepath.suffix.lower()
        if filetype != ".csv":
            raise TypeError("Currently, only .csv files are supported")
        csv_data = open(datasource, mode="r")
    elif isinstance(datasource, bytes):
        # Convert bytes to a string using io.StringIO to simulate a file
        csv_data = io.StringIO(datasource.decode("utf-8"), newline="")
    else:
        raise TypeError(f"File type {type(datasource)} not valid")
    return csv_data


def check_savefile(filename, ext):
    """
    enforce that a savefile has the correct extension
    """

    if not ext.startswith("."):
        ext = "." + ext

    if isinstance(filename, str):
        if not filename.lower().endswith(ext):
            filename += ext
    elif isinstance(filename, pathlib.PosixPath):
        if not filename.suffix == ext:
            filename = filename.parent / (filename.name + ext)
    return filename


def validate_spectra(spectra, required_keys=None):
    """check that a spectra passed not-from-source"""

    # first, spectra must be a dict
    if not isinstance(spectra, dict):
        raise TypeError("Must be dict")

    # check any required keys
    if required_keys is not None:
        if isinstance(required_keys, list):
            for key in required_keys:
                validate_key(key, spectra)
        elif isinstance(required_keys, str):
            validate_key(required_keys, spectra)

    # check that all values within a dict are the same length
    if len(np.unique([len(val) for key, val in spectra.items()])) > 1:
        raise ValueError("All entries in the spectra dict must be of the same length")

    return spectra


def validate_key(key, dct):
    if key not in dct:
        raise KeyError("Required key {key} is absent.")


def rows_to_bytes(rows, encoding="cp1252"):
    buffer = io.StringIO()
    writer = csv.writer(buffer)
    writer.writerows(rows)

    # Get the CSV data from buffer, convert to bytes
    csv_data = buffer.getvalue()
    csv_bytes = csv_data.encode(encoding)  # encode to bytes
    return csv_bytes


def fig_to_bytes(fig):
    buf = io.BytesIO()
    fig.savefig(
        buf, format="png"
    )  # You can change the format as needed (e.g., 'jpeg', 'pdf')
    buf.seek(0)  # Rewind the buffer
    return buf.getvalue()


def get_lamp_positions(num_lamps, x, y, num_divisions=100):
    """
    generate a list of (x,y) positions for a lamp given room dimensions and
    the number of lamps desired
    """
    lst = [new_lamp_position(i + 1, x, y) for i in range(num_lamps)]
    return np.array(lst).T


def new_lamp_position(lamp_idx, x, y, num_divisions=100):
    """
    get the default position for an additional new lamp
    x and y are the room dimensions
    first index is 1, not 0.
    """
    xp = np.linspace(0, x, num_divisions + 1)
    yp = np.linspace(0, y, num_divisions + 1)
    xidx, yidx = _get_idx(lamp_idx, num_divisions=num_divisions)
    return xp[xidx], yp[yidx]


def _get_idx(num_points, num_divisions=100):
    grid_size = (num_divisions, num_divisions)
    return _place_points(grid_size, num_points)[-1]


def _place_points(grid_size, num_points):
    M, N = grid_size
    grid = np.zeros(grid_size)
    points = []

    # Place the first point in the center
    center = (M // 2, N // 2)
    points.append(center)
    grid[center] = 1  # Marking the grid cell as occupied

    for _ in range(1, num_points):
        max_dist = -1
        best_point = None

        for x in range(M):
            for y in range(N):
                if grid[x, y] == 0:
                    # Calculate the minimum distance to all existing points
                    min_point_dist = min(
                        [np.sqrt((x - px) ** 2 + (y - py) ** 2) for px, py in points]
                    )
                    # Calculate the distance to the nearest boundary
                    min_boundary_dist = min(x, M - 1 - x, y, N - 1 - y)
                    # Find the point where the minimum of these distances is maximized
                    min_dist = min(min_point_dist, min_boundary_dist)

                    if min_dist > max_dist:
                        max_dist = min_dist
                        best_point = (x, y)

        if best_point:
            points.append(best_point)
            grid[best_point] = 1  # Marking the grid cell as occupied
    return points



import warnings
from pathlib import Path
import datetime
import inspect
import json
import zipfile
import io
import numpy as np
from collections.abc import Iterable
from .lamp import Lamp
from .calc_zone import CalcZone, CalcPlane, CalcVol
from ._data import get_version
from ._helpers import parse_loadfile, check_savefile, fig_to_bytes
from .room_plotter import RoomPlotter
from .disinfection_calculator import DisinfectionCalculator
from .reflectance import ReflectanceManager

VALID_UNITS = ["meters", "feet"]


class Room:
    """
    Represents a room containing lamps and calculation zones.

    The room is defined by its dimensions and can contain multiple lamps and
    different calculation zones for evaluating lighting conditions and other
    characteristics.
    """

    def __init__(
        self,
        x=None,
        y=None,
        z=None,
        units=None,
        standard=None,
        reflectances=None,
        reflectance_x_spacings=None,
        reflectance_y_spacings=None,
        max_num_passes=None,
        air_changes=None,
        ozone_decay_constant=None,
    ):
        self.units = "meters" if units is None else units.lower()
        if self.units not in VALID_UNITS:
            raise KeyError(f"Invalid unit {units}")
        default_dimensions = (
            [6.0, 4.0, 2.7] if self.units == "meters" else [20.0, 13.0, 9.0]
        )
        self.x = default_dimensions[0] if x is None else x
        self.y = default_dimensions[1] if y is None else y
        self.z = default_dimensions[2] if z is None else z

        self.standard = (
            "ANSI IES RP 27.1-22 (ACGIH Limits)" if standard is None else standard
        )
        self.ozone_decay_constant = (
            2.7 if ozone_decay_constant is None else ozone_decay_constant
        )
        self.air_changes = 1.0 if air_changes is None else air_changes

        self.ref_manager = ReflectanceManager(
            self.x,
            self.y,
            self.z,
            reflectances,
            reflectance_x_spacings,
            reflectance_y_spacings,
            max_num_passes,
        )
        self.plotter = RoomPlotter(self)
        self.disinfection = DisinfectionCalculator(self)

        self.set_dimensions()

        self.lamps = {}
        self.calc_zones = {}
        self.calc_state = {}
        self.update_state = {}

    def set_max_num_passes(self, max_num_passes):
        self.ref_manager.max_num_passes = max_num_passes
        return self

    def set_reflectance(self, R, wall_id=None):
        self.ref_manager.set_reflectance(R=R, wall_id=wall_id)
        return self

    def set_reflectance_spacing(self, x_spacing=None, y_spacing=None, wall_id=None):
        self.ref_manager.set_spacing(
            x_spacing=x_spacing, y_spacing=y_spacing, wall_id=wall_id
        )
        return self

    def to_dict(self):
        data = {}
        data["x"] = self.x
        data["y"] = self.y
        data["z"] = self.z
        data["units"] = self.units
        data["reflectances"] = self.ref_manager.reflectances
        data["reflectance_x_spacing"] = self.ref_manager.x_spacings
        data["reflectance_y_spacing"] = self.ref_manager.y_spacings
        data["max_num_passes"] = self.ref_manager.max_num_passes
        data["standard"] = self.standard
        data["air_changes"] = self.air_changes
        data["ozone_decay_constant"] = self.ozone_decay_constant

        dct = self.__dict__.copy()
        data["lamps"] = {k: v.save_lamp() for k, v in dct["lamps"].items()}
        data["calc_zones"] = {k: v.save_zone() for k, v in dct["calc_zones"].items()}
        return data

    def save(self, fname=None):
        """save all relevant parameters to a json file"""
        savedata = {}
        version = get_version(Path(__file__).parent / "_version.py")
        savedata["guv-calcs_version"] = version

        now = datetime.datetime.now()
        now_local = datetime.datetime.now(now.astimezone().tzinfo)
        timestamp = now_local.strftime("%Y-%m-%d %H:%M:%S %Z%z")
        savedata["timestamp"] = timestamp

        savedata["data"] = self.to_dict()

        if fname is not None:
            filename = check_savefile(fname, ".guv")
            with open(filename, "w") as json_file:
                json.dump(savedata, json_file, indent=4)
        else:
            return json.dumps(savedata, indent=4)

    @classmethod
    def load(cls, filedata):
        """load from a file"""

        load_data = parse_loadfile(filedata)

        saved_version = load_data["guv-calcs_version"]
        current_version = get_version(Path(__file__).parent / "_version.py")
        if saved_version != current_version:
            warnings.warn(
                f"This file was saved with guv-calcs {saved_version}, while you have {current_version} installed."
            )
        room_dict = load_data["data"]

        roomkeys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        room = cls(**{k: v for k, v in room_dict.items() if k in roomkeys})

        for lampid, lamp in room_dict["lamps"].items():
            room.add_lamp(Lamp.from_dict(lamp))

        for zoneid, zone in room_dict["calc_zones"].items():
            if zone["calctype"] == "Plane":
                room.add_calc_zone(CalcPlane.from_dict(zone))
            elif zone["calctype"] == "Volume":
                room.add_calc_zone(CalcVol.from_dict(zone))
        return room

    def export_zip(
        self,
        fname=None,
        include_plots=False,
        include_lamp_files=False,
        include_lamp_plots=False,
    ):
        """
        write the project file and all results files to a zip file. Optionally include
        extra files like lamp ies files, spectra files, and plots.
        """

        # save project
        data_dict = {"room.guv": self.save()}

        # save all results
        for zone_id, zone in self.calc_zones.items():
            if zone.calctype != "Zone":
                data_dict[zone.name + ".csv"] = zone.export()
                if include_plots:
                    if zone.calctype == "Plane":
                        # Save the figure to a BytesIO object
                        if zone.dose:
                            title = f"{zone.hours} Hour Dose ({zone.height} m)"
                        else:
                            title = f"Irradiance ({zone.height} m)"
                        fig, ax = zone.plot_plane(title=title)
                        data_dict[zone.name + ".png"] = fig_to_bytes(fig)

        for lamp_id, lamp in self.lamps.items():
            if lamp.filedata is not None:
                if include_lamp_files:
                    data_dict[lamp.name + ".ies"] = lamp.save_ies()
                if include_lamp_plots:
                    ies_fig, ax = lamp.plot_ies(title=lamp.name)
                    data_dict[lamp.name + "_ies.png"] = fig_to_bytes(ies_fig)
            if lamp.spectra is not None:
                if include_lamp_plots:
                    linfig, _ = lamp.spectra.plot(
                        title=lamp.name, yscale="linear", weights=True, label=True
                    )
                    logfig, _ = lamp.spectra.plot(
                        title=lamp.name, yscale="log", weights=True, label=True
                    )
                    linkey = lamp.name + "_spectra_linear.png"
                    logkey = lamp.name + "_spectra_log.png"
                    data_dict[linkey] = fig_to_bytes(linfig)
                    data_dict[logkey] = fig_to_bytes(logfig)

        zip_buffer = io.BytesIO()
        # Create a zip file within this BytesIO object
        with zipfile.ZipFile(zip_buffer, "w", zipfile.ZIP_DEFLATED) as zip_file:
            # Loop through the dictionary, adding each string/byte stream to the zip
            for filename, content in data_dict.items():
                # Ensure the content is in bytes
                if isinstance(content, str):
                    content = content.encode("utf-8")
                # Add the file to the zip; writing the bytes to a BytesIO object for the file
                file_buffer = io.BytesIO(content)
                zip_file.writestr(filename, file_buffer.getvalue())
        zip_bytes = zip_buffer.getvalue()

        if fname is not None:
            with open(fname, "wb") as f:
                f.write(zip_bytes)
        else:
            return zip_bytes

    def get_calc_state(self):
        """
        Save all the features in the room that, if changed, will require re-calculation
        """

        room_state = [  # temp..this should be put with the ref_manager eventually
            self.ref_manager.x_spacings.copy(),
            self.ref_manager.y_spacings.copy(),
        ]

        lamp_state = {}
        for key, lamp in self.lamps.items():
            if lamp.enabled:
                lamp_state[key] = lamp.get_calc_state()

        zone_state = {}
        for key, zone in self.calc_zones.items():
            if zone.calctype != "Zone" and zone.enabled:
                zone_state[key] = zone.get_calc_state()

        calc_state = {}
        calc_state["room"] = room_state
        calc_state["lamps"] = lamp_state
        calc_state["calc_zones"] = zone_state

        return calc_state

    def get_update_state(self):
        """
        Save all the features in the room that should NOT trigger
        a recalculation, only an update
        """

        room_state = [
            self.ref_manager.reflectances.copy(),
            self.units,
        ]

        lamp_state = {}
        for key, lamp in self.lamps.items():
            lamp_state[key] = lamp.get_update_state()

        zone_state = {}
        for key, zone in self.calc_zones.items():
            if zone.calctype != "Zone":
                zone_state[key] = zone.get_update_state()

        update_state = {}
        update_state["room"] = room_state
        update_state["lamps"] = lamp_state
        update_state["calc_zones"] = zone_state

        return update_state

    def set_units(self, units):
        """set room units"""
        if units not in ["meters", "feet"]:
            raise KeyError("Valid units are `meters` or `feet`")
        self.units = units
        self._update_standard_zones() 
        self.harmonize_units()
        return self

    def harmonize_units(self):
        """ensure that all lamps in the state have the correct units"""
        for lamp in self.lamps.values():
            if lamp.surface.units != self.units:
                lamp.set_units(self.units)

    def set_dimensions(self, x=None, y=None, z=None):
        """set room dimensions"""
        self.x = self.x if x is None else x
        self.y = self.y if y is None else y
        self.z = self.z if z is None else z
        self.dimensions = (self.x, self.y, self.z)
        self.volume = self.x * self.y * self.z

        self.ref_manager.update_dimensions(self.x, self.y, self.z)
        return self

    def get_units(self):
        """return room units"""
        return self.units

    def get_dimensions(self):
        """return room dimensions"""
        self.dimensions = np.array((self.x, self.y, self.z))
        return self.dimensions

    def get_volume(self):
        """return room volume"""
        self.volume = self.x * self.y * self.z
        return self.volume

    def get_disinfection_data(self, zone_id="WholeRoomFluence"):
        """return the fluence_dict, dataframe, and violin plot"""
        return self.disinfection.get_disinfection_data(zone_id=zone_id)
        
    def _update_standard_zones(self):
        """update the standard safety calculation zones based on the current standard and units"""
        if "UL8802" in self.standard:
            height = 1.9 if self.units == "meters" else 6.25
            skin_horiz = False
            eye_vert = False
            fov_vert = 180
        else:
            height = 1.8 if self.units == "meters" else 5.9
            skin_horiz = True
            eye_vert = True
            fov_vert = 80
            
        if "SkinLimits" in self.calc_zones.keys():
            self.calc_zones["SkinLimits"].set_height(height)
            self.calc_zones["SkinLimits"].horiz = skin_horiz
        if "EyeLimits" in self.calc_zones.keys():
            self.calc_zones["EyeLimits"].set_height(height)
            self.calc_zones["EyeLimits"].fov_vert = fov_vert
            self.calc_zones["EyeLimits"].vert = eye_vert

    def set_standard(self, standard):
        """update the photobiological safety standard the Room is subject to"""
        self.standard = standard
        self._update_standard_zones()        
        return self

    def add_standard_zones(self):
        """
        convenience function. Add skin and eye limit calculation planes,
        plus whole room fluence.
        """

        max_vol_val = 20
        max_plane_val = 50

        self.add_calc_zone(
            CalcVol(
                zone_id="WholeRoomFluence",
                name="Whole Room Fluence",
                x1=0,
                x2=self.x,
                y1=0,
                y2=self.y,
                z1=0,
                z2=self.z,
                num_x=min(int(self.x * 20), max_vol_val),
                num_y=min(int(self.y * 20), max_vol_val),
                num_z=min(int(self.z * 20), max_vol_val),
                show_values=False,
            )
        )

        self.add_calc_zone(
            CalcPlane(
                zone_id="SkinLimits",
                name="Skin Dose (8 Hours)",
                x1=0,
                x2=self.x,
                y1=0,
                y2=self.y,
                num_x=min(int(self.x * 20), max_plane_val),
                num_y=min(int(self.y * 20), max_plane_val),
                dose=True,
                hours=8,
            )
        )

        self.add_calc_zone(
            CalcPlane(
                zone_id="EyeLimits",
                name="Eye Dose (8 Hours)",
                x1=0,
                x2=self.x,
                y1=0,
                y2=self.y,
                num_x=min(int(self.x * 20), max_plane_val),
                num_y=min(int(self.y * 20), max_plane_val),
                horiz=False,
                fov_horiz=180,
                dose=True,
                hours=8,
            )
        )
        
        # sets the height and field of view parameters
        self._update_standard_zones()
        
        return self

    def _check_position(self, dimensions, obj_name):
        """
        Method to check if an object's dimensions exceed the room's boundaries.
        """
        msg = None
        for coord, roomcoord in zip(dimensions, self.dimensions):
            if coord > roomcoord:
                msg = f"{obj_name} exceeds room boundaries!"
                warnings.warn(msg, stacklevel=2)
        return msg

    def check_lamp_position(self, lamp):
        return self._check_position(lamp.position, lamp.name)

    def check_zone_position(self, calc_zone):
        if isinstance(calc_zone, CalcPlane):
            dimensions = [calc_zone.x2, calc_zone.y2]
        elif isinstance(calc_zone, CalcVol):
            dimensions = [calc_zone.x2, calc_zone.y2, calc_zone.z2]
        elif isinstance(calc_zone, CalcZone):
            # this is a hack; a generic CalcZone is just a placeholder
            dimensions = self.dimensions
        return self._check_position(dimensions, calc_zone.name)

    def check_positions(self):
        msgs = []
        for lamp_id, lamp in self.lamps.items():
            msgs.append(self.check_lamp_position(lamp))
        for zone_id, zone in self.calc_zones.items():
            msgs.append(self.check_zone_position(zone))
        return msgs

    def add_lamp(self, lamp):
        """
        Adds a lamp to the room if it fits within the room's boundaries.
        """
        if not isinstance(lamp, Lamp):
            raise TypeError(f"Must be type Lamp, not {type(lamp)}")
        # check units
        if lamp.surface.units != self.units:
            lamp.set_units(self.units)
        # check position
        self.check_lamp_position(lamp)
        self.lamps[lamp.lamp_id] = lamp
        return self

    def remove_lamp(self, lamp_id):
        """remove a lamp from the room"""
        del self.lamps[lamp_id]
        return self

    def add_calc_zone(self, calc_zone):
        """
        Adds a calculation zone to the room if it fits within the room's boundaries.
        """
        if not isinstance(calc_zone, (CalcZone, CalcPlane, CalcVol)):
            raise TypeError(
                f"Must be CalcZone, CalcPlane, or CalcVol not {type(calc_zone)}"
            )
        self.check_zone_position(calc_zone)
        self.calc_zones[calc_zone.zone_id] = calc_zone
        return self

    def add(self, *args):
        """
        Add objects to the Room.

        - If an object is a Lamp, it is added as a lamp.
        - If an object is a CalcZone, CalcPlane, or CalcVol, it is added as a calculation zone.
        - If an object is iterable, it is recursively processed.
        - Otherwise, a warning is printed.
        """

        for obj in args:
            if isinstance(obj, Lamp):
                self.add_lamp(obj)
            elif isinstance(obj, (CalcZone, CalcPlane, CalcVol)):
                self.add_calc_zone(obj)
            elif isinstance(obj, dict):
                self.add(*obj.values())
            elif isinstance(obj, Iterable) and not isinstance(obj, (str, bytes)):
                self.add(*obj)  # Recursively process other iterables
            else:
                msg = f"Cannot add object of type {type(obj).__name__} to Room."
                warnings.warn(msg, stacklevel=3)

        return self

    def remove_calc_zone(self, zone_id):
        """remove calculation zone from room"""
        del self.calc_zones[zone_id]
        return self

    def _get_valid_lamps(self):
        """return"""
        return {
            k: v for k, v in self.lamps.items() if v.enabled and v.filedata is not None
        }

    def calculate(self, hard=False):
        """
        Triggers the calculation of lighting values in each calculation zone
        based on the current lamps in the room.

        If no updates have been made since the last calculate call that would
        require a full recalculation, either only an update will be performed
        or no recalculation will occur.

        If `hard` is True, this behavior is overriden and the full
        recalculation will be performed
        """

        self.harmonize_units()

        valid_lamps = self._get_valid_lamps()

        if len(valid_lamps) > 0:

            new_calc_state = self.get_calc_state()
            new_update_state = self.get_update_state()

            LAMP_RECALC = self.calc_state.get("lamps") != new_calc_state.get("lamps")
            REF_RECALC = self.calc_state.get("room") != new_calc_state.get("room")
            REF_UPDATE = self.update_state.get("room") != new_update_state.get("room")

            # calculate incidence on the surfaces if the reflectances or lamps have changed
            if LAMP_RECALC or REF_RECALC or REF_UPDATE or hard:
                self.ref_manager.calculate_incidence(valid_lamps, hard=hard)

            for name, zone in self.calc_zones.items():
                if zone.enabled and len(valid_lamps) > 0:
                    zone.calculate_values(
                        lamps=valid_lamps, ref_manager=self.ref_manager, hard=hard
                    )
            # update calc states.
            self.calc_state = new_calc_state
            self.update_state = new_update_state

            # possibly this should be per-calc zone? idk maybe it's fine
            for lamp_id in valid_lamps.keys():
                new_calc_state = self.lamps[lamp_id].get_calc_state()
                self.lamps[lamp_id].calc_state = new_calc_state
        else:
            if len(self.lamps) == 0:
                msg = "No lamps are present in the room."
            else:
                msg = "No valid lamps are present in the room--either lamps have been disabled, or filedata has not been provided."
            warnings.warn(msg, stacklevel=3)

        return self

    def calculate_by_id(self, zone_id, hard=False):
        """calculate just the calc zone selected"""
        valid_lamps = {
            k: v for k, v in self.lamps.items() if v.enabled and v.filedata is not None
        }
        self.calc_zones[zone_id].calculate_values(lamps=valid_lamps)
        self.calc_state = self.get_calc_state()
        self.update_state = self.get_update_state()
        return self

    def plotly(self, fig=None, select_id=None, title=""):
        """return a plotly figure of all the room's components"""
        self.harmonize_units()
        return self.plotter.plotly(fig=fig, select_id=select_id, title=title)



import csv
import warnings
from importlib import resources
import numpy as np
import matplotlib.pyplot as plt
from ._helpers import load_csv, rows_to_bytes


class Spectrum:
    """
    Attributes:
        wavelengths: 1d arraylike
            wavelength values in nanometers
        intensities: 1d arraylike
            relative intensity values in arbitrary units
        weights: dict
            Dictionary containing spectral weights. Keys are labels for the weighting
            table and values are a tuple of two lists, (wavelengths, intensities).
            Generally loaded from within the package but you can pass your own if you
            really want to.
    """

    def __init__(self, wavelengths, intensities, weights=None):
        self.raw_wavelengths = np.array(wavelengths)
        self.raw_intensities = np.array(intensities)

        # sort values in case they are out of order
        sortidx = np.argsort(self.raw_wavelengths)
        self.raw_wavelengths = self.raw_wavelengths[sortidx]
        self.raw_intensities = self.raw_intensities[sortidx]

        if len(self.raw_wavelengths) != len(self.raw_intensities):
            raise ValueError("Number of wavelengths and intensities do not match.")

        # store copy for later retrieval
        self.wavelengths = self.raw_wavelengths
        self.intensities = self.raw_intensities

        self.weights = self._load_spectral_weightings() if weights is None else weights
        self.weighted_intensities = (
            None if self.weights is None else self._weight_spectra()
        )

    @classmethod
    def from_file(cls, filepath):
        """
        initialize spectrum object from file
        Be careful initializing from file, as the class assumes the file's
        first column correponds to 'Wavelengths' and the second column
        corresponds to 'Intensities'. 'Wavelengths' should be in units of
        nanometers, 'Intensities' should be in units of relative intensity.
        """
        csv_data = load_csv(filepath)
        reader = csv.reader(csv_data)
        # read each line
        spectra = []
        for i, row in enumerate(reader):
            try:
                vals = list(map(float, row))
                spectra.append((vals[0], vals[1]))
            except ValueError:
                if i == 0:  # probably a header
                    continue
                # else:
                # warnings.warn(f"Skipping invalid datarow: {row}")
        if len(spectra) == 0:
            raise ValueError("File contains no valid data.")

        wavelengths = np.array(spectra).T[0]
        intensities = np.array(spectra).T[1]
        return cls(wavelengths, intensities)

    @classmethod
    def from_dict(cls, dct):
        """
        initialize spectrum object from a dictionary where
        the first key contains wavelength values, and the second
        key contains intensity values
        """
        keys = list(dct.keys())
        wavelengths = np.array(dct[keys[0]])
        intensities = np.array(dct[keys[1]])
        return cls(wavelengths, intensities)

    def to_dict(self, as_string=False):
        spec = {}
        spec["Wavelength"] = self.wavelengths
        spec["Unweighted Relative Intensity"] = self.intensities
        if self.weighted_intensities is not None:
            for key, val in self.weighted_intensities.items():
                spec[key] = val
        if as_string:
            for key, val in spec.items():
                spec[key] = ", ".join(map(str, spec[key]))
        return spec

    def to_csv(self, fname=None):
        rows = [list(self.to_dict().keys())]
        vals = [self.wavelengths, self.intensities]
        vals += list(self.weighted_intensities.values())
        rows += list(np.array(vals).T)
        csv_bytes = rows_to_bytes(rows)
        if fname is not None:
            with open(fname, "wb") as csvfile:
                csvfile.write(csv_bytes)
        else:
            return csv_bytes

    def _log_interp(self, wvs, weight_wvs, weight_values):
        """log10 interpolation for the tlv weights"""
        logterp = np.interp(wvs, weight_wvs, np.log10(weight_values))
        return np.power(10, logterp)

    def _load_spectral_weightings(self):
        """load spectral weightings from within package"""

        fname = "UV Spectral Weighting Curves.csv"
        path = resources.files("guv_calcs.data").joinpath(fname)
        with path.open("rb") as file:
            weights = file.read()

        csv_data = load_csv(weights)
        reader = csv.reader(csv_data, delimiter=",")
        headers = next(reader, None)  # get headers

        data = {}
        for header in headers:
            data[header] = []
        for row in reader:
            for header, value in zip(headers, row):
                data[header].append(float(value))

        spectral_weightings = {}
        for i, (key, val) in enumerate(data.items()):
            if i == 0:
                wavelengths = np.array(val)
            else:
                spectral_weightings[key] = (wavelengths, np.array(val))
        return spectral_weightings

    def weight(self, standard):
        """weight by spectral effectiveness"""
        if standard in self.weights.keys():
            weight_wavelengths, weights_orig = self.weights[standard]
            weights = self._log_interp(
                self.wavelengths, weight_wavelengths, weights_orig
            )
            weighted_intensity = self.intensities * weights
        else:
            warnings.warn(f"{standard} not in available weightings.")
        return weighted_intensity

    def _weight_spectra(self):
        """calculate the weighted spectra"""
        weighted_intensities = {}
        maxval = max(self.intensities)
        for key, val in self.weights.items():
            weighted_intensity = self.weight(key)
            ratio = maxval / max(weighted_intensity)
            weighted_intensities[key] = weighted_intensity * ratio
        return weighted_intensities

    def _scale(self, value):
        """scale both the base and weighted intensities by a value"""
        self.intensities *= value
        if self.weighted_intensities is not None:
            for key, val in self.weighted_intensities.items():
                self.weighted_intensities[key] *= value

    def _filter(self, wavelength, intensity, minval=None, maxval=None):
        """internal method for filtering. Returns tuple (wavelength,intensity)"""
        if minval is None:
            minval = min(self.wavelengths)
        if maxval is None:
            maxval = max(self.wavelengths)
        idx1 = np.argwhere(wavelength >= minval)
        idx2 = np.argwhere(wavelength <= maxval)
        idx = np.intersect1d(idx1, idx2)
        return wavelength[idx], intensity[idx]

    def revert(self):
        """restore wavelength and intensity values to original values"""
        self.wavelengths = self.raw_wavelengths
        self.intensities = self.raw_intensities
        return self

    def filter(self, minval=None, maxval=None):
        """
        filter the spectra and weighted spectra over a wavelength range
        """
        for key, val in self.weighted_intensities.items():
            w2, i2 = self._filter(self.wavelengths, val, minval, maxval)
            self.weighted_intensities[key] = i2
        w, i = self._filter(self.wavelengths, self.intensities, minval, maxval)
        self.wavelengths = w
        self.intensities = i
        return self

    def sum(self, minval=None, maxval=None, weight=None):
        if weight in self.weighted_intensities.keys():
            intensities = self.weighted_intensities[weight]
            wavelengths, intensities = self._filter(
                self.wavelengths, intensities, minval, maxval
            )
        else:
            wavelengths, intensities = self._filter(
                self.wavelengths, self.intensities, minval, maxval
            )
        return sum_spectrum(wavelengths, intensities)

    def scale(self, power, minval=None, maxval=None):
        """
        scale the spectra to a power value, such that the total spectral
        power is equal to the value.
        optionally, consider only the power output over a range of wavelengths
        """
        spectral_power = self.sum(minval=minval, maxval=maxval)
        self._scale(power / spectral_power)
        return self

    def normalize(self, normval=1):
        """normalize the maximum intensity to a value"""
        self._scale(normval / max(self.intensities))
        return self

    def get_tlv(self, standard):
        """return the total uv dose over 8 hours for this specific lamp,
        per a particular spectral effectiveness standard. units: mJ/cm2"""
        weights = self._log_interp(
            self.wavelengths, *self.weights[standard]
        )  # get weights
        i_new = self.intensities / self.sum()  # scale
        s_lambda = sum_spectrum(self.wavelengths, weights * i_new)
        return 3 / s_lambda

    def get_max_irradiance(self, standard):
        """return max irradiance for this specific spectra, per a particular
        spectral effectiveness standard. units: uW/cm2"""
        return self.get_tlv(standard) / 60 / 60 / 8 * 1000

    def get_seconds_to_tlv(self, irradiance, standard):
        """
        determine the number of seconds before a TLV is reached
        provided an irradiance value in uW/cm2
        and a spectral effectiveness standard
        """
        weights = self._log_interp(
            self.wavelengths, *self.weights[standard]
        )  # get weights
        i_new = self.intensities * irradiance / self.sum()  # scale
        weighted_irradiance = sum_spectrum(self.wavelengths, weights * i_new)
        return 3000 / weighted_irradiance

    def plot(
        self,
        title="",
        fig=None,
        ax=None,
        figsize=(6.4, 4.8),
        yscale="linear",
        label=None,
        weights=False,
    ):
        """
        plot the spectra and any weighted spectra.
        `yscale` is generally either "linear" or "log", but any matplotlib
        scale is permitted
        label: str or bool or None
            if str, the spectrum will be labeled with that str. If True,
            default label is `Unweighted Relative Intensity`. If False, label
            will not be used. if
        weights: bool or str
            If not False or None,
        """

        if fig is None:
            if ax is None:
                fig, ax = plt.subplots()
            else:
                fig = plt.gcf()
        else:
            if ax is None:
                ax = fig.axes[0]

        default_label = "Unweighted Relative Intensity"
        if label:
            if isinstance(label, str):
                ax.plot(self.wavelengths, self.intensities, label=label)
            else:
                ax.plot(self.wavelengths, self.intensities, label=default_label)
        else:
            ax.plot(self.wavelengths, self.intensities)

        if weights:
            if isinstance(weights, str):
                if weights in self.weighted_intensities.keys():
                    val = self.weighted_intensities[weights]
                    ax.plot(
                        self.wavelengths,
                        self.weighted_intensities[weights],
                        label=weights,
                        alpha=0.7,
                        linestyle="--",
                    )
            elif isinstance(weights, bool):
                for key, val in self.weighted_intensities.items():
                    ax.plot(self.wavelengths, val, label=key, alpha=0.7, linestyle="--")

        if label or weights:
            ax.legend()
        ax.grid(True, which="both", ls="--", c="gray", alpha=0.3)
        ax.set_xlabel("Wavelength [nm]")
        ax.set_ylabel("Relative intensity [%]")
        ax.set_yscale(yscale)
        ax.set_title(title)
        return fig, ax


def sum_spectrum(wavelength, intensity):
    """
    sum across a spectrum.
    ALWAYS use this when summing across a spectra!!!
    """
    weighted_intensity = [
        intensity[i] * (wavelength[i] - wavelength[i - 1])
        for i in range(1, len(wavelength))
    ]
    return sum(weighted_intensity)



import numpy as np
from photompy import get_intensity
from .trigonometry import attitude, to_polar
from ._units import convert_units


class LightingCalculator:
    """
    Performs all computations for a calculation zone
    """

    def __init__(self, zone):
        self.zone = zone

    def compute(self, lamps, hard=False):
        """
        Calculate and return irradiance values at all coordinate points within the zone.
        """

        # this will only recalculate if the lamp has changed--unless it is a 'hard' recalculation
        self.zone.lamp_values_base = {
            lamp_id: self._calculate_lamp(lamp, hard) for lamp_id, lamp in lamps.items()
        }

        # this step is always cheap
        self.zone.lamp_values = {
            lamp_id: self._apply_filters(lamps[lamp_id], values.copy())
            for lamp_id, values in self.zone.lamp_values_base.items()
        }

        # sum the base lamp values
        if self.zone.fov_horiz < 360 and len(lamps) > 1:
            values = self._calculate_horizontal_fov(lamps)
        else:
            values = sum(self.zone.lamp_values.values())

        # reshape
        values = values.reshape(*self.zone.num_points)

        return values

    def _calculate_lamp(self, lamp, hard=False):
        """
        Calculate the zone values for a single lamp
        """

        NEW_LAMP = self.zone.lamp_values_base.get(lamp.lamp_id) is None
        LAMP_UPDATE = lamp.calc_state != lamp.get_calc_state()
        ZONE_UPDATE = self.zone.calc_state != self.zone.get_calc_state()

        if hard or NEW_LAMP or LAMP_UPDATE or ZONE_UPDATE:
            # get coords
            rel_coords = self.zone.coords - lamp.surface.position
            Theta, Phi, R = self._transform_lamp_coords(rel_coords, lamp)
            if lamp.surface.units.lower() != "meters":
                R = np.array(convert_units(lamp.surface.units, "meters", *R))
            # fetch intensity values from photometric data
            interpdict = lamp.lampdict["interp_vals"]
            values = get_intensity(Theta, Phi, interpdict) / R ** 2

            # near field only if necessary
            if lamp.surface.source_density > 0 and lamp.surface.photometric_distance:
                values = self._calculate_nearfield(lamp, R, values)

            if np.isnan(values.any()):  # mask any nans near light source
                values = np.ma.masked_invalid(values)

        else:
            values = self.zone.lamp_values_base[lamp.lamp_id]

        return values

    def _apply_filters(self, lamp, values):
        """
        update the values of a single lamp based on the calc zone properties,
        but which don't require a full recalculation
        """

        rel_coords = self.zone.coords - lamp.surface.position
        Theta0, Phi0, R0 = to_polar(*rel_coords.T)
        # apply vertical field of view
        values[Theta0 < 90 - self.zone.fov_vert / 2] = 0

        if self.zone.vert:
            values *= np.sin(np.radians(Theta0))
        if self.zone.horiz:
            values *= abs(np.cos(np.radians(Theta0)))

        if lamp.intensity_units.lower() == "mw/sr":
            values = values / 10  # convert from mW/Sr to uW/cm2

        # reshape
        values = values.reshape(*self.zone.num_points)

        return values

    def _calculate_nearfield(self, lamp, R, values):
        """
        calculate the values within the photometric distance
        over a discretized source
        """
        near_idx = np.where(R < lamp.surface.photometric_distance)
        # set current values to zero
        values[near_idx] = 0
        # redo calculation in a loop
        num_points = len(lamp.surface.surface_points)
        for point, val in zip(
            lamp.surface.surface_points, lamp.surface.intensity_map.reshape(-1)
        ):
            rel_coords = self.zone.coords - point
            Theta, Phi, R = self._transform_lamp_coords(rel_coords, lamp)
            Theta_n, Phi_n, R_n = Theta[near_idx], Phi[near_idx], R[near_idx]
            interpdict = lamp.lampdict["interp_vals"]
            if lamp.surface.units.lower() != "meters":
                R_n = np.array(convert_units(lamp.surface.units, "meters", *R_n))
            near_values = get_intensity(Theta_n, Phi_n, interpdict) / R_n ** 2
            near_values = near_values * val / num_points
            values[near_idx] += near_values
        return values

    def _calculate_horizontal_fov(self, lamps):
        """
        Vectorized function to compute the largest possible value for all lamps
        within a horizontal view field.
        """

        # Compute relative coordinates: Shape (num_points, num_lamps, 3)
        lamp_positions = np.array(
            [lamp.surface.position for lamp in lamps.values() if lamp.enabled]
        )
        rel_coords = self.zone.coords[:, None, :] - lamp_positions[None, :, :]

        # Calculate horizontal angles (in degrees)
        angles = np.degrees(np.arctan2(rel_coords[..., 1], rel_coords[..., 0]))
        angles = angles % 360  # Wrap; Shape (N, M)
        angles = angles[:, :, None]  # Expand; Shape (N, M, 1)

        # Compute pairwise angular differences for all rows
        diffs = np.abs(angles - angles.transpose(0, 2, 1))
        diffs = np.minimum(diffs, 360 - diffs)  # Wrap angular differences to [0, 180]

        # Create the adjacency mask for each pair within 180 degrees
        adjacency = diffs <= self.zone.fov_horiz / 2  # Shape (N, M, M)

        # current values to be transformed
        vals = self.zone.lamp_values.values()
        values = np.array([val.reshape(-1) for val in vals]).T
        # Sum the values for all connected components (using the adjacency mask)
        value_sums = adjacency @ values[:, :, None]  # Shape (N, M, 1)
        # Remove the last singleton dimension,
        value_sums = value_sums.squeeze(-1)  # Shape (N, M)

        return np.max(value_sums, axis=1)  # Shape (N,)

    def _transform_lamp_coords(self, rel_coords, lamp):
        """
        transform zone coordinates to be consistent with any lamp
        transformations applied, and convert to polar coords for further
        operations
        """
        Theta0, Phi0, R0 = to_polar(*rel_coords.T)
        # apply all transformations that have been applied to this lamp, but in reverse
        rel_coords = np.array(
            attitude(rel_coords.T, roll=0, pitch=0, yaw=-lamp.heading)
        ).T
        rel_coords = np.array(attitude(rel_coords.T, roll=0, pitch=-lamp.bank, yaw=0)).T
        rel_coords = np.array(
            attitude(rel_coords.T, roll=0, pitch=0, yaw=-lamp.angle)
        ).T
        Theta, Phi, R = to_polar(*rel_coords.T)
        return Theta, Phi, R
